#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for converging a kkr calculation and
some helper methods to do so with AiiDA
"""
import numpy as np
from masci_tools.io.kkr_params import kkrparams
from masci_tools.io.common_functions import get_Ry2eV, get_ef_from_potfile
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.common.exceptions import NotExistentAttributeError
from aiida.engine import (
    WorkChain,
    while_,
    if_,
    ToContext,
    workfunction,
    calcfunction,
)
from aiida_kkr.calculations.kkr import KkrCalculation
from aiida_kkr.calculations.voro import VoronoiCalculation
from aiida_kkr.tools.extract_kkrhost_noco_angles import extract_noco_angles
from aiida_kkr.tools.common_workfunctions import (
    test_and_get_codenode,
    get_inputs_kkr,
    get_parent_paranode,
    update_params_wf,
)
from aiida_kkr.workflows.voro_start import kkr_startpot_wc
from aiida_kkr.workflows.dos import kkr_dos_wc

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.11.2'
__contributors__ = (u'Jens Broeder', u'Philipp Rüßmann', u'David Antognini Silva')

eV2Ry = 1.0 / get_Ry2eV()

# TODO: magnetism (init and converge magnetic state)
# TODO: check convergence (RMAX, GMAX etc.)
# TODO: save timing info of the steps
# TODO: switch to LLOYD
# TODO: emin-emax setting
# TODO: restart from workflow output instead of calculation output
# TODO: add warnings
# TODO: maybe define the energy point density instead of a fixed number as in input?
# TODO: overwrite defaults from parent if parent is previous kkr_scf run
# TODO: retrieve DOS within scf run


class kkr_scf_wc(WorkChain):
    """
    Workchain for converging a KKR calculation (SCF).

    It converges the charge potential.
    Two paths are possible:

    (1) Start from a structure and run a voronoi calculation first,
    optional with calc_parameters
    (2) Start from an existing Voronoi or KKR calculation, with a remoteData

    :param wf_parameters: (Dict), Workchain Specifications
    :param options: (Dict); specifications for the computer
    :param structure: (StructureData), Crystal structure
    :param calc_parameters: (Dict), Voronoi/Kkr Parameters
    :param remote_data: (RemoteData), from a KKR, or Voronoi calculation
    :param voronoi: (Code)
    :param kkr: (Code)

    :return output_kkr_scf_wc_para: (Dict), Information of workflow results
        like Success, last result node, list with convergence behavior

    minimum input example:
    1. Code1, Code2, Structure, (Parameters), (wf_parameters)
    2. Code2, remote_data, (Parameters), (wf_parameters)

    maximum input example:
    1. Code1, Code2, Structure, Parameters
        wf_parameters: {'queue_name' : String,
                        'resources' : dict({"num_machines": int, "num_mpiprocs_per_machine" : int})
                        'walltime' : int}
    2. Code2, (remote-data), wf_parameters as in 1.

    Hints:
    1. This workflow does not work with local codes!
    """

    _workflowversion = __version__

    # default workflow parameters
    _wf_default = AttributeDict()
    # Maximum number of kkr jobs/starts (defauld iterations per start)
    _wf_default.kkr_runmax = 5
    # Stop if charge density is converged below this value
    _wf_default.convergence_criterion = 10**-8
    # reduce mixing factor by this factor if calculation fails due to too large mixing
    _wf_default.mixreduce = 0.5
    # temperature increasing steps if workflow fails
    _wf_default.tempr_increase = 50.
    # threshold after which agressive mixing is used
    _wf_default.threshold_aggressive_mixing = 8 * 10**-3
    _wf_default.strmix = 0.03  # mixing factor of simple mixing
    _wf_default.brymix = 0.05  # mixing factor of aggressive mixing
    # number of iterations done per KKR calculation
    _wf_default.nsteps = 50
    # wether or not pre-convergnce on a coarse energy and k-point grid should be done
    _wf_default.coarse_preconvergence = True
    # convergence settings for coarse preconvergence
    _wf_default.convergence_setting_coarse = AttributeDict()  # setting of the coarse preconvergence
    _wf_default.convergence_setting_coarse.npol = 7
    _wf_default.convergence_setting_coarse.n1 = 3
    _wf_default.convergence_setting_coarse.n2 = 11
    _wf_default.convergence_setting_coarse.n3 = 3
    _wf_default.convergence_setting_coarse.tempr = 1000.0
    _wf_default.convergence_setting_coarse.kmesh = [10, 10, 10]
    # threshold after which final conversion settings are used
    _wf_default.threshold_switch_high_accuracy = 10**-3
    # setting of the final convergence (lower tempr, 48 epts, denser k-mesh)
    _wf_default.convergence_setting_fine = AttributeDict()
    _wf_default.convergence_setting_fine.npol = 5
    _wf_default.convergence_setting_fine.n1 = 7
    _wf_default.convergence_setting_fine.n2 = 29
    _wf_default.convergence_setting_fine.n3 = 7
    _wf_default.convergence_setting_fine.tempr = 600.0
    _wf_default.convergence_setting_fine.kmesh = [30, 30, 30]
    # initialize and converge magnetic calculation
    _wf_default.mag_init = False
    # external magnetic field used in initialization step in Ry
    _wf_default.hfield = 0.02
    # position in unit cell where magnetic field is applied [default (None) means apply to all]
    _wf_default.init_pos = None
    # fix direction of magnetic moment if the direction changes less
    # than this value in degrees (calculated as sqrt((delta theta)**2 + (delta phi)**2))
    _wf_default.fix_dir_threshold = 1.0
    # add DOS to testopts and retrieve dos.atom files in each scf run
    _wf_default.retreive_dos_data_scf_run = False
    # minimal distance of start of the energy contour to highest lying core state in Ry
    _wf_default.delta_e_min_core_states = 0.2  # Ry
    # set these keys from defaults in kkr_startpot workflow since they are only passed onto that workflow
    for key, value in kkr_startpot_wc.get_wf_defaults(silent=True).items():
        if key in [
            'dos_params',
            'fac_cls_increase',
            'natom_in_cls_min',
            'delta_e_min',
            'threshold_dos_zero',
            'check_dos',
        ]:
            _wf_default[key] = value

    # default options
    _options_default = AttributeDict()
    _options_default.queue_name = ''  # Queue name to submit jobs too
    # resources to allowcate for the job
    _options_default.resources = AttributeDict()
    _options_default.resources.num_machines = 1
    # walltime after which the job gets killed (gets parsed to KKR)
    _options_default.max_wallclock_seconds = 60 * 60
    _options_default.withmpi = True  # execute KKR with mpi or without
    _options_default.custom_scheduler_commands = ''  # some additional scheduler commands

    # intended to guide user interactively in setting up a valid wf_params node
    @classmethod
    def get_wf_defaults(cls, silent=False):
        """Print and return _wf_default dictionary.

        Can be used to easily create set of wf_parameters.
        returns _wf_default, _options_default
        """
        if not silent:
            print(f'Version of workflow: {cls._workflowversion}')
        return cls._wf_default, cls._options_default.copy()

    @classmethod
    def define(cls, spec):
        """Defines the outline of the workflow."""
        # Take input of the workflow or use defaults defined above
        super(kkr_scf_wc, cls).define(spec)
        spec.input(
            'wf_parameters',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._wf_default),
            help="""
            Settings for the workflow. Use `KkrCalculation.get_wf_defaults()`
            to get the default values and default options.
            """
        )
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=False,
            default=lambda: orm.Dict(dict=cls._wf_default),
            help="""
            Computer settings used by the calculations in the workflow
            (see also help string of wf_parameters).
            """
        )
        spec.input(
            'structure',
            valid_type=orm.StructureData,
            required=False,
            help="""
            Input structure for which a calculation is started with a
            VoronoiCalculation.
            Can be skipped if a previous KkrCalculation is given with the
            `remote_data` input node.
            """
        )
        spec.input(
            'calc_parameters',
            valid_type=orm.Dict,
            required=False,
            help="""
            KKR-specific calculation parameters (LMAX etc.),
            usually set up with the help of the `kkrparams` class.
            """
        )
        spec.input(
            'remote_data',
            valid_type=orm.RemoteData,
            required=False,
            help="""
            RemoteFolder node of a preconverged calculation.
            Can be used as a starting point to skip the Voronoi step.
            """
        )
        spec.input(
            'voronoi',
            valid_type=orm.Code,
            required=False,
            help="""
            Voronoi code node, needed only if `structure` input node is given.
            """
        )
        spec.input(
            'kkr',
            valid_type=orm.Code,
            required=True,
            help='KKRhost code node which will run the KkrCalculations',
        )
        spec.input(
            'startpot_overwrite',
            valid_type=orm.SinglefileData,
            required=False,
            help="""
            Potential SinglefileData, can be used to overwrite
            the starting potential from Voronoi
            (the shapefun will be used though and thus needs to be compatible).
            This can be used to construct a better starting potential from a
            preconverged calculation (e.g. in a smaller unit cell).
            """
        )
        spec.input(
            'initial_noco_angles',
            valid_type=orm.Dict,
            required=False,
            help="""
            Initial non-collinear angles for the magnetic moments of the
            impurities. See KkrCalculation for details.
            """
        )
        spec.input(
            'params_kkr_overwrite',
            valid_type=orm.Dict,
            required=False,
            help='Set some input parameters of the KKR calculation.'
        )

        # define output nodes
        spec.output(
            'output_kkr_scf_wc_ParameterResults',
            valid_type=orm.Dict,
            required=True,
        )
        spec.output(
            'last_calc_out',
            valid_type=orm.Dict,
            required=False,
        )
        spec.output(
            'last_RemoteData',
            valid_type=orm.RemoteData,
            required=False,
        )
        spec.output(
            'last_InputParameters',
            valid_type=orm.Dict,
            required=False,
        )
        spec.output(
            'results_vorostart',
            valid_type=orm.Dict,
            required=False,
        )
        spec.output(
            'starting_dosdata_interpol',
            valid_type=orm.XyData,
            required=False,
        )
        spec.output(
            'final_dosdata_interpol',
            valid_type=orm.XyData,
            required=False,
        )
        spec.output(
            'last_noco_angles',
            valid_type=orm.Dict,
            required=False,
        )

        # Here the structure of the workflow is defined
        spec.outline(
            cls.start,
            # check if voronoi run needed, otherwise skip this step
            if_(cls.validate_input)(
                # run kkr_startpot workflow (sets up voronoi input, runs voro calc, does some consistency checks)
                cls.run_voronoi,
                # check output of run_voronoi and determine if calculation has to be terminated here
                cls.check_voronoi
            ),
            # while loop for KKR run(s), first simple mixing
            # then Anderson with corase pre-convergence settings
            # finally convergence step with higher accuracy
            while_(cls.condition)(
                # update parameters for kkr step using previous output(s)
                cls.update_kkr_params,
                # run kkr step
                # TODO: encapsulate this in restarting mechanism
                # (should be a base class of workflows that start calculations)
                # i.e. use base_restart_calc workchain as parent
                cls.run_kkr,
                # check results for convergence and collect some intermediate results
                cls.inspect_kkr
            ),
            # compute final dos if check_dos is True
            cls.get_dos,
            cls.check_dos,
            # finalize calculation and create output nodes
            cls.return_results
        )

        # definition of exit codes if the workflow needs to be terminated
        spec.exit_code(
            221,
            'ERROR_NO_PARENT_PARAMS_FOUND',
            message='Unable to extract parent paremeter node of input remote folder',
        )
        spec.exit_code(
            222,
            'ERROR_INVALID_KKR_CODE',
            message='The code you provided for kkr does not use the plugin kkr.kkr',
        )
        spec.exit_code(
            223,
            'ERROR_INVALID_VORONOI_CODE',
            message='The code you provided for voronoi does not use the plugin kkr.voro',
        )
        spec.exit_code(
            224,
            'ERROR_NO_VORONOI_CODE_GIVEN',
            message='ERROR: StructureData was provided, but no voronoi code was provided',
        )
        spec.exit_code(
            225,
            'ERROR_NOT_ENOUGH_INPUTS',
            message='ERROR: No StructureData nor remote_data was provided as Input',
        )
        spec.exit_code(
            226,
            'ERROR_KKR_STARTPOT_FAILED',
            message='ERROR: kkr_startpot_wc step failed!',
        )
        spec.exit_code(
            227,
            'ERROR_DOS_RUN_UNSUCCESSFUL',
            message='DOS run unsuccessful. Check inputs.',
        )
        spec.exit_code(
            228,
            'ERROR_CALC_PARAMETERS_INCOMPLETE',
            message='ERROR: calc_parameters given are not consistent! Missing mandatory keys',
        )
        spec.exit_code(
            229,
            'ERROR_CALC_PARAMTERS_INCONSISTENT',
            message='ERROR: calc_parameters given are not consistent! Hint: did you give an unknown keyword?',
        )
        spec.exit_code(
            230,
            'ERROR_NO_CALC_PARAMETERS_GIVEN',
            message='ERROR: calc_parameters not given as input but are needed!',
        )
        spec.exit_code(
            231,
            'ERROR_PARAM_UPDATE_FAILED',
            message='ERROR: parameter update unsuccessful: some key, value pair not valid!',
        )
        spec.exit_code(
            232,
            'ERROR_CALC_PARAMTERS_INCOMPLETE',
            message='ERROR: calc_parameters misses keys',
        )
        spec.exit_code(
            233,
            'ERROR_LAST_REMOTE_NOT_FOUND',
            message='ERROR: last_remote could not be set to a previous successful calculation',
        )
        spec.exit_code(
            234,
            'ERROR_MAX_KKR_RESTARTS_REACHED',
            message='ERROR: maximal number of KKR restarts reached. Exiting now!',
        )
        spec.exit_code(
            235,
            'ERROR_CALC_SUBMISSION_FAILED',
            message='ERROR: last KKRcalc in SUBMISSIONFAILED state',
        )

    def start(self):
        """Initialize context and some parameters."""
        self.report(f'INFO: started KKR convergence workflow version {self._workflowversion}')

        ####### init #######

        # internal para /control para
        self.ctx.loop_count = 0
        self.ctx.last_mixing_scheme = 0
        self.ctx.calcs = []
        self.ctx.abort = False
        # flags used internally to check whether the individual steps were successful
        self.ctx.dos_ok = False
        self.ctx.voro_step_success = False
        self.ctx.kkr_step_success = False
        self.ctx.kkr_converged = False
        self.ctx.kkr_higher_accuracy = False
        # links to previous calculations
        self.ctx.last_calc = None
        self.ctx.last_params = None
        self.ctx.last_remote = None
        # convergence info about rms etc. (used to determine convergence behavior)
        self.ctx.last_rms_all = []
        self.ctx.rms_all_steps = []
        self.ctx.last_neutr_all = []
        self.ctx.neutr_all_steps = []

        # input para
        wf_dict = self.inputs.wf_parameters.get_dict()
        options_dict = self.inputs.options.get_dict()

        if wf_dict == {}:
            wf_dict = self._wf_default
            self.report('INFO: using default wf parameter')
        if options_dict == {}:
            options_dict = self._options_default
            self.report('INFO: using default options')

        # set values from input, or defaults
        self.ctx.withmpi = options_dict.get(
            'withmpi',
            self._options_default['withmpi'],
        )
        self.ctx.resources = options_dict.get(
            'resources',
            self._options_default['resources'],
        )
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds',
            self._options_default['max_wallclock_seconds'],
        )
        self.ctx.queue = options_dict.get(
            'queue_name',
            self._options_default['queue_name'],
        )
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands',
            self._options_default['custom_scheduler_commands'],
        )
        self.ctx.options_params_dict = orm.Dict(
            dict={
                'withmpi': self.ctx.withmpi,
                'resources': self.ctx.resources,
                'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
                'queue_name': self.ctx.queue,
                'custom_scheduler_commands': self.ctx.custom_scheduler_commands
            }
        )

        # set label and description
        self.ctx.description_wf = self.inputs.get(
            'description', 'Workflow for '
            'a KKR scf calculation starting '
            'either from a structure with '
            'automatic voronoi calculation '
            'or a valid RemoteData node of '
            'a previous calculation'
        )
        self.ctx.label_wf = self.inputs.get('label', 'kkr_scf_wc')

        # set workflow parameters
        self.ctx.max_number_runs = wf_dict.get(
            'kkr_runmax',
            self._wf_default['kkr_runmax'],
        )
        self.ctx.strmix = wf_dict.get(
            'strmix',
            self._wf_default['strmix'],
        )
        self.ctx.brymix = wf_dict.get(
            'brymix',
            self._wf_default['brymix'],
        )
        self.ctx.check_dos = wf_dict.get(
            'check_dos',
            self._wf_default['check_dos'],
        )
        self.ctx.dos_params = wf_dict.get(
            'dos_params',
            self._wf_default['dos_params'],
        )
        self.ctx.convergence_criterion = wf_dict.get(
            'convergence_criterion',
            self._wf_default['convergence_criterion'],
        )
        self.ctx.coarse_preconvergence = wf_dict.get(
            'coarse_preconvergence',
            self._wf_default['coarse_preconvergence'],
        )
        self.ctx.convergence_setting_coarse = wf_dict.get(
            'convergence_setting_coarse',
            self._wf_default['convergence_setting_coarse'],
        )
        self.ctx.convergence_setting_fine = wf_dict.get(
            'convergence_setting_fine',
            self._wf_default['convergence_setting_fine'],
        )
        self.ctx.mixreduce = wf_dict.get(
            'mixreduce',
            self._wf_default['mixreduce'],
        )
        self.ctx.tempr_increase = wf_dict.get(
            'tempr_increase',
            self._wf_default['tempr_increase'],
        )
        self.ctx.nsteps = wf_dict.get(
            'nsteps',
            self._wf_default['nsteps'],
        )
        self.ctx.threshold_aggressive_mixing = wf_dict.get(
            'threshold_aggressive_mixing',
            self._wf_default['threshold_aggressive_mixing'],
        )
        self.ctx.threshold_switch_high_accuracy = wf_dict.get(
            'threshold_switch_high_accuracy',
            self._wf_default['threshold_switch_high_accuracy'],
        )

        # initial magnetization
        self.ctx.mag_init = wf_dict.get(
            'mag_init',
            self._wf_default['mag_init'],
        )
        self.ctx.hfield = wf_dict.get(
            'hfield',
            self._wf_default['hfield'],
        )
        self.ctx.xinit = wf_dict.get(
            'init_pos',
            self._wf_default['init_pos'],
        )
        self.ctx.mag_init_step_success = False

        # difference in eV to emin (e_fermi) if emin (emax) are larger
        # (smaller) than emin (e_fermi)
        self.ctx.delta_e = wf_dict.get(
            'delta_e_min',
            self._wf_default['delta_e_min'],
        )
        # threshold for dos comparison (comparison of dos at emin)
        self.ctx.threshold_dos_zero = wf_dict.get(
            'threshold_dos_zero',
            self._wf_default['threshold_dos_zero'],
        )
        # distance of EMIN to highest core state (used to modify EMIN start)
        self.ctx.delta_e_min_core_states = wf_dict.get(
            'delta_e_min_core_states',
            self._wf_default['delta_e_min_core_states'],
        )

        # this will be used to store the Fermi level
        self.ctx.efermi = None

        # set starting noco angles, gets updated in between the KKR runs if
        # fix_dir == False for all atoms
        if 'initial_noco_angles' in self.inputs:
            self.ctx.initial_noco_angles = self.inputs.initial_noco_angles
            self.ctx.fix_dir_threshold = wf_dict.get(
                'fix_dir_threshold',
                self._wf_default['fix_dir_threshold'],
            )

        # retreive dos data in each scf run
        self.ctx.scf_dosdata = wf_dict.get(
            'retreive_dos_data_scf_run',
            self._wf_default['retreive_dos_data_scf_run'],
        )

        self.report(
            'INFO: use the following parameter:\n'
            '\nGeneral settings\n'
            f'use mpi: {self.ctx.withmpi}\n'
            f'max number of KKR runs: {self.ctx.max_number_runs}\n'
            f'Resources: {self.ctx.resources}\n'
            f'Walltime (s): {self.ctx.max_wallclock_seconds}\n'
            f'queue name: {self.ctx.queue}\n'
            f'scheduler command: {self.ctx.custom_scheduler_commands}\n'
            f'description: {self.ctx.description_wf}\n'
            f'label: {self.ctx.label_wf}\n'
            f'\nMixing parameter\n'
            f'Straight mixing factor: {self.ctx.strmix}\n'
            f'Anderson mixing factor: {self.ctx.brymix}\n'
            f'Nsteps scf cycle: {self.ctx.nsteps}\n'
            f'Convergence criterion: {self.ctx.convergence_criterion}\n'
            f'threshold_aggressive_mixing: {self.ctx.threshold_aggressive_mixing}\n'
            f'threshold_switch_high_accuracy: {self.ctx.threshold_switch_high_accuracy}\n'
            f'use coarse preconvergence: {self.ctx.coarse_preconvergence}\n'
            f'convergence_setting_coarse: {self.ctx.convergence_setting_coarse}\n'
            f'convergence_setting_fine: {self.ctx.convergence_setting_fine}\n'
            f'factor reduced mixing if failing calculation: {self.ctx.mixreduce}\n'
            f'temperature increasing value if failing calculation: {self.ctx.tempr_increase}\n'
            f'\nAdditional parameter\n'
            f'check DOS between runs: {self.ctx.check_dos}\n'
            f'DOS parameters: {self.ctx.dos_params}\n'
            f'init magnetism in first step: {self.ctx.mag_init}\n'
            f'init magnetism, hfield: {self.ctx.hfield}\n'
            f'init magnetism, init_pos: {self.ctx.xinit}\n'
        )

        self.ctx.successful = True
        self.ctx.rms = []
        self.ctx.neutr = []
        self.ctx.warnings = []
        self.ctx.errors = []
        self.ctx.formula = ''

        # for results table each list gets one entry per iteration that has been performed
        self.ctx.KKR_steps_stats = {
            'success': [],
            'isteps': [],
            'imix': [],
            'mixfac': [],
            'qbound': [],
            'high_sett': [],
            'first_rms': [],
            'last_rms': [],
            'first_neutr': [],
            'last_neutr': [],
            'pk': [],
            'uuid': []
        }

    def validate_input(self):
        """Validate input and find out which path (1, or 2) to take

        return True means run voronoi if false run kkr directly
        """
        run_voronoi = True
        inputs = self.inputs

        if 'structure' in inputs:
            self.report('INFO: Found structure in input. Start with Voronoi calculation.')
            if not 'voronoi' in inputs:
                return self.exit_codes.ERROR_NO_VORONOI_CODE_GIVEN  # pylint: disable=no-member
        elif 'remote_data' in inputs:
            self.report('INFO: Found remote_data in input. Continue calculation without running voronoi step.')
            run_voronoi = False
        else:
            return self.exit_codes.ERROR_NOT_ENOUGH_INPUTS  # pylint: disable=no-member

        if 'voronoi' in inputs:
            try:
                test_and_get_codenode(inputs.voronoi, 'kkr.voro', use_exceptions=True)
            except ValueError:
                return self.exit_codes.ERROR_INVALID_VORONOI_CODE  # pylint: disable=no-member

        if 'kkr' in inputs:
            try:
                test_and_get_codenode(inputs.kkr, 'kkr.kkr', use_exceptions=True)
            except ValueError:
                return self.exit_codes.ERROR_INVALID_KKR_CODE  # pylint: disable=no-member

        # set params and remote folder to input if voronoi step is skipped
        if not run_voronoi:
            self.ctx.last_remote = inputs.remote_data
            num_parents = len(self.ctx.last_remote.get_incoming(node_class=orm.CalcJobNode).all_link_labels())
            if num_parents == 0:
                pk_last_remote = self.ctx.last_remote.inputs.last_RemoteData.outputs.output_kkr_scf_wc_ParameterResults.get_dict(
                ).get('last_calc_nodeinfo').get('pk')
                last_calc = orm.load_node(pk_last_remote)
                self.ctx.last_remote = last_calc.outputs.remote_folder
            try:  # first try parent of remote data output of a previous calc.
                parent_params = get_parent_paranode(inputs.remote_data)
            except AttributeError:
                try:  # next try to extract parameter from previous kkr_scf_wc output
                    parent_params = inputs.remote_data.inputs.last_RemoteData.inputs.calc_parameters
                except AttributeError:
                    return self.exit_codes.ERROR_NO_PARENT_PARAMS_FOUND  # pylint: disable=no-member
            if 'calc_parameters' in inputs:
                self.ctx.last_params = inputs.calc_parameters
                # TODO: check last_params consistency against parent_params
                #parent_params_dict = parent_params.get_dict()
                # for key, val in self.ctx.last_params.get_dict().iteritems():
                #    if key in parent_params_dict.keys():
                #        if val != parent_params_dict[key]:
                #
                #    else:
                #
            else:
                self.ctx.last_params = parent_params
            self.ctx.voro_step_success = True

        return run_voronoi

    def run_voronoi(self):
        """Run the voronoi step calling voro_start workflow."""

        # collects inputs
        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        voronoicode = self.inputs.voronoi
        kkrcode = self.inputs.kkr

        # set KKR parameters if any are given, otherwise use defaults
        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None

        # check if default values are missing and set appropriately from defaults
        defaults, version = kkrparams.get_KKRcalc_parameter_defaults()
        if params is None:
            params = orm.Dict(dict=defaults)
            params.label = 'default values'
        newparams = {}
        for key, val in defaults.items():
            # this one is automatically set by kkr_startpot_wc so we skip it here
            if params.get_dict().get(key, None) is None and key != 'RCLUSTZ':
                newparams[key] = val
                self.report(f'INFO: Automatically added default values to KKR parameters: {key} {val}')
        if newparams != {}:
            for key, val in params.get_dict().items():
                if val is not None:
                    newparams[key] = val
            updatenode = orm.Dict(dict=newparams)
            updatenode.label = 'added defaults to KKR input parameter'
            updatenode.description = 'Overwritten KKR input parameter to correct missing default values automatically'
            params = update_params_wf(params, updatenode)

        # set nspin to 2 if mag_init is used
        if self.ctx.mag_init:
            input_dict = params.get_dict()
            para_check = kkrparams()
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)
            nspin_in = para_check.get_value('NSPIN')
            if nspin_in is None:
                nspin_in = 1
            if nspin_in < 2:
                self.report('WARNING: found NSPIN=1 but for maginit needs NPIN=2. Overwrite this automatically')
                para_check.set_value('NSPIN', 2, silent=True)
                self.report(f'INFO: update parameters to: {para_check.get_set_values()}')
                updatenode = orm.Dict(dict=para_check.get_dict())
                updatenode.label = 'overwritten KKR input parameters'
                updatenode.description = 'Overwritten KKR input parameter to correct NSPIN to 2'
                params = update_params_wf(params, updatenode)

        # check consistency of input parameters before running calculation
        self.check_input_params(params, is_voronoi=True)

        # set parameters of voro_start sub workflow
        sub_wf_params_dict = kkr_startpot_wc.get_wf_defaults(silent=True)
        label, description = 'voro_start_default_params', 'workflow parameters for voro_start'
        if 'wf_parameters' in self.inputs:
            wf_params_input = self.inputs.wf_parameters.get_dict()
            num_updated = 0
            for key in list(sub_wf_params_dict.keys()):
                if key in list(wf_params_input.keys()):
                    val = wf_params_input[key]
                    sub_wf_params_dict[key] = val
                    num_updated += 1
            if num_updated > 0:
                label = 'voro_start_updated_params'
        sub_wf_params = orm.Dict(dict=sub_wf_params_dict)
        sub_wf_params.label = label
        sub_wf_params.description = description

        self.report('INFO: run voronoi step')
        self.report(f'INFO: using calc_params ({params})')
        self.report(f'INFO: using wf_parameters ({sub_wf_params})')

        wf_label = 'kkr_startpot (voronoi)'
        wf_desc = 'subworkflow to set up the input of a KKR calculation'
        builder = kkr_startpot_wc.get_builder()
        builder.kkr = kkrcode
        builder.voronoi = voronoicode
        builder.calc_parameters = params
        builder.wf_parameters = sub_wf_params
        builder.structure = structure
        builder.metadata.label = wf_label  # pylint: disable=no-member
        builder.metadata.description = wf_desc  # pylint: disable=no-member
        builder.options = self.ctx.options_params_dict
        if 'startpot_overwrite' in self.inputs:
            builder.startpot_overwrite = self.inputs.startpot_overwrite
        future = self.submit(builder)

        return ToContext(voronoi=future, last_calc=future)

    def check_voronoi(self):
        """Check output of kkr_startpot_wc workflow.

        It checks the starting potential, shapefun etc.
        """

        self.report('INFO: checking voronoi output')
        voro_step_ok = False

        # check some output
        kkrstartpot_results = self.ctx.voronoi.outputs.results_vorostart_wc.get_dict()
        if kkrstartpot_results['successful']:
            voro_step_ok = True
        # initialize last_remote and last_params (gets updated in loop of KKR calculations)
        self.ctx.last_params = self.ctx.voronoi.outputs.last_params_voronoi
        self.ctx.last_remote = self.ctx.voronoi.outputs.last_voronoi_remote

        # store result in context
        self.ctx.voro_step_success = voro_step_ok

        # abort calculation if something failed in voro_start step
        if not voro_step_ok:
            return self.exit_codes.ERROR_KKR_STARTPOT_FAILED  # pylint: disable=no-member

        self.report('INFO: done checking voronoi output')
        return None

    def condition(self):
        """
        check convergence condition
        """
        self.report('INFO: checking condition for kkr step')
        do_kkr_step = True
        stopreason = ''

        # increment KKR runs loop counter
        self.ctx.loop_count += 1

        # check if initial step was successful
        if not self.ctx.voro_step_success:
            stopreason = 'voronoi step unsucessful'
            do_kkr_step = False
            return do_kkr_step

        # check if previous calculation reached convergence criterion
        if self.ctx.kkr_converged:
            if not self.ctx.kkr_higher_accuracy:
                do_kkr_step = do_kkr_step & True
            else:
                stopreason = 'KKR converged'
                do_kkr_step = False
        else:
            do_kkr_step = do_kkr_step & True

        # next check only needed if another iteration should be done after validating convergence etc. (previous checks)
        if do_kkr_step:
            # check if maximal number of iterations has been reached
            if self.ctx.loop_count <= self.ctx.max_number_runs:
                do_kkr_step = do_kkr_step & True
            else:
                do_kkr_step = False
                # TODO do this differently, return of exit code needs to be done outside of while loop!
                # return self.exit_codes.ERROR_MAX_KKR_RESTARTS_REACHED
                return do_kkr_step

        self.report(f'INFO: done checking condition for kkr step (result={do_kkr_step})')

        if not do_kkr_step:
            self.report(f'INFO: stopreason={stopreason}')

        return do_kkr_step

    def update_kkr_params(self):
        """Update set of KKR parameters.

        (check for reduced mixing, change of mixing strategy, change of accuracy setting)
        """
        self.report('INFO: updating kkr param step')
        decrease_mixing_fac = False
        switch_agressive_mixing = False
        switch_higher_accuracy = False
        initial_settings = False

        # only do something other than sample mixing after first kkr run
        if self.ctx.loop_count != 1:
            # first determine if previous step was successful
            # (otherwise try to find some rms value and decrease mixing to try again)
            if not self.ctx.kkr_step_success:
                try:
                    # check if calculation did start (maybe cluster had some hiccup)
                    calc = self.ctx.calcs[-1]
                    has_output_node = len(calc.get_outgoing(link_label_filter='output_parameters').all()) > 0
                    self.report(
                        'INFO: last KKR calculation failed. '
                        'Probably because the cluster had some issue. '
                        'Try to resubmit the same calculation'
                    )
                except:
                    # otherwise try to decrease the mixing factor
                    decrease_mixing_fac = True
                    self.report('INFO: last KKR calculation failed. Trying to decrease mixfac')

            convergence_on_track = self.convergence_on_track()

            # check if calculation was on its way to converge
            if not convergence_on_track:
                decrease_mixing_fac = True
                self.report('INFO: last KKR did not converge. trying decreasing mixfac')
                self.report(f'INFO: ctx.calcs: {self.ctx.calcs} {type(self.ctx.calcs)}')
                # reset last_remote to last successful calculation
                # needs to be list because `(x)range` does not support slicing
                calclist = list(range(len(self.ctx.calcs)))
                if len(calclist) > 1:
                    calclist = calclist[::-1]  # go backwards through list
                for icalc in calclist:
                    self.report(f'INFO: last calc success? {icalc} {self.ctx.KKR_steps_stats["success"][icalc]}')
                    if self.ctx.KKR_steps_stats['success'][icalc]:
                        self.ctx.last_remote = self.ctx.calcs[icalc].outputs.remote_folder
                        break  # exit loop if last_remote was found successfully
                    self.ctx.last_remote = None
                # if no previous calculation was successful take voronoi output
                # or remote data from input (depending on the inputs)
                self.report(f'INFO: last_remote is None? {self.ctx.last_remote is None} {"structure" in self.inputs}')
                if self.ctx.last_remote is None:
                    if 'structure' in self.inputs:
                        self.ctx.last_remote = self.ctx.voronoi.outputs.last_voronoi_remote
                    else:
                        self.ctx.last_remote = self.inputs.remote_data
                # check if last_remote has finally been set and abort if this is not the case
                self.report(f'INFO: last_remote is still None? {self.ctx.last_remote is None}')
                if self.ctx.last_remote is None:
                    return self.exit_codes.ERROR_LAST_REMOTE_NOT_FOUND  # pylint: disable=no-member

            # check if mixing strategy should be changed
            last_mixing_scheme = self.ctx.last_params.get_dict()['IMIX']
            if last_mixing_scheme is None:
                last_mixing_scheme = 0

            if convergence_on_track:
                last_rms = self.ctx.last_rms_all[-1]

                if (last_rms < self.ctx.threshold_aggressive_mixing and last_mixing_scheme == 0):
                    switch_agressive_mixing = True
                    self.report('INFO: rms low enough, switch to agressive mixing')

                # check if switch to higher accuracy should be done
                if not self.ctx.kkr_higher_accuracy and last_rms < self.ctx.threshold_switch_high_accuracy:
                    switch_higher_accuracy = True
                    self.report('INFO: rms low enough, switch to higher accuracy settings')
        else:
            initial_settings = True

        # check if coarse pre-convergence is done
        if not self.ctx.coarse_preconvergence:
            switch_higher_accuracy = True

        # if needed update parameters
        if (
            decrease_mixing_fac or switch_agressive_mixing or switch_higher_accuracy or initial_settings or
            self.ctx.mag_init
        ):

            if initial_settings:
                label = 'initial KKR scf parameters'
                description = 'initial parameter set for scf calculation'
            else:
                label = ''
                description = ''

            # step 1: extract info from last input parameters and check consistency
            params = self.ctx.last_params
            input_dict = params.get_dict()
            para_check = kkrparams()

            # step 1.1: try to fill keywords
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)

            # step 1.2: check if all mandatory keys are there and add defaults if missing
            new_params = self.initialize_params(para_check)

            # step 2: change parameter (contained in new_params dictionary)

            # adapt EMIN if core states are too close
            new_params = self.adapt_emin(new_params)

            # adapt mixing and convergence settings
            new_params, label, description = self.change_conv_para(
                new_params, para_check, initial_settings, decrease_mixing_fac, switch_agressive_mixing,
                switch_higher_accuracy, label, description
            )

            # initial magnetization
            new_params = self.initial_mag(new_params, initial_settings)

            # step 2.2 update values
            try:
                for key, val in new_params.items():
                    para_check.set_value(key, val, silent=True)
            except:
                return self.exit_codes.ERROR_PARAM_UPDATE_FAILED  # pylint: disable=no-member

            # step 3:
            self.report(f'INFO: update parameters to: {para_check.get_set_values()}')
            updatenode = orm.Dict(dict=para_check.get_dict())
            updatenode.label = label
            updatenode.description = description

            paranode_new = update_params_wf(self.ctx.last_params, updatenode)
            self.ctx.last_params = paranode_new
        else:
            self.report('INFO: reuse old settings')

        self.report('INFO: done updating kkr param step')

        return None

    def initialize_params(self, para_check):
        """Initialize new_params with missing defaults"""
        # init new_params dict where updated params are collected
        new_params = {}

        # find and add missing defaults
        missing_list = para_check.get_missing_keys(use_aiida=True)
        if missing_list != []:
            kkrdefaults = kkrparams.get_KKRcalc_parameter_defaults()[0]
            kkrdefaults_updated = []
            for key_default, val_default in list(kkrdefaults.items()):
                if key_default in missing_list:
                    new_params[key_default] = kkrdefaults.get(key_default)
                    kkrdefaults_updated.append(key_default)
            if len(kkrdefaults_updated) > 0:
                self.report(f'ERROR: calc_parameters misses keys: {missing_list}')
                return self.exit_codes.ERROR_CALC_PARAMTERS_INCOMPLETE  # pylint: disable=no-member
            self.report(f'updated KKR parameter node with default values: {kkrdefaults_updated}')

        return new_params

    def adapt_emin(self, new_params):
        """Change EMIN if core states are too close."""

        # get parent calculation (either Voronoi or KkrCalculation)
        parent_calc = self.ctx.last_remote.get_incoming(node_class=orm.CalcJobNode).first().node

        # get highest lying core state
        parent_calc_out = parent_calc.outputs.output_parameters
        ecores = parent_calc_out['core_states_group']['energy_highest_lying_core_state_per_atom']
        ecore_max = max(list(set([ec for ec in ecores if ec is not None])))

        # get emin from Voronoi or KkrCalculation
        if parent_calc.process_class == VoronoiCalculation:
            emin_old = parent_calc.outputs.output_parameters['emin']
        else:
            emin_old = parent_calc.outputs.output_parameters['energy_contour_group']['emin']

        # find new emin if distance to core states is too small
        ecore_mindist = self.ctx.delta_e_min_core_states  # in Ry
        if abs(ecore_max - emin_old) < ecore_mindist:
            # move emin lower by delta_e (converted to Ry units)
            emin_new = ecore_max - self.ctx.delta_e * eV2Ry
            self.report(
                f'INFO: Core states too close to start of contour.'
                f'\nChanging EMIN to {emin_new} (ecore_max={ecore_max}, emin_old={emin_old})'
            )
            new_params['EMIN'] = emin_new

        return new_params

    def change_conv_para(
        self, new_params, para_check, initial_settings, decrease_mixing_fac, switch_agressive_mixing,
        switch_higher_accuracy, label, description
    ):
        """Adapt the kkr parameters to change the convergence settings and the mixing"""

        if initial_settings and 'structure' in self.inputs:
            # make sure to ignore IMIX from input node
            # (start with simple mixing even if IMIX is set otherwise)
            # this is enforced whenever voronoi step is starting point
            # (otherwise you may want to continue a preconverged calculation)
            last_mixing_scheme = None
        else:
            last_mixing_scheme = para_check.get_value('IMIX')
        if last_mixing_scheme is None:
            last_mixing_scheme = 0

        strmixfac = self.ctx.strmix
        brymixfac = self.ctx.brymix
        nsteps = self.ctx.nsteps

        # add number of scf steps
        new_params['NSTEPS'] = nsteps

        # step 2.1 fill new_params dict with values to be updated
        if decrease_mixing_fac:
            if last_mixing_scheme == 0:
                self.report(f'(strmixfax, mixreduce)= ({strmixfac}, {self.ctx.mixreduce})')
                self.report(f'type(strmixfax, mixreduce)= {type(strmixfac)} {type(self.ctx.mixreduce)}')
                strmixfac = strmixfac * self.ctx.mixreduce
                self.ctx.strmix = strmixfac
                label += f'decreased_mix_fac_str (step {self.ctx.loop_count})'
                description += f'decreased STRMIX factor by {self.ctx.mixreduce}'
            else:
                self.report(f'(brymixfax, mixreduce)= ({brymixfac}, {self.ctx.mixreduce})')
                self.report(f'type(brymixfax, mixreduce)= {type(brymixfac)} {type(self.ctx.mixreduce)}')
                brymixfac = brymixfac * self.ctx.mixreduce
                self.ctx.brymix = brymixfac
                label += 'decreased_mix_fac_bry'
                description += f'decreased BRYMIX factor by {self.ctx.mixreduce}'

        # add mixing factor
        new_params['STRMIX'] = strmixfac
        new_params['BRYMIX'] = brymixfac

        if switch_agressive_mixing:
            last_mixing_scheme = 5
            label += ' switched_to_agressive_mixing'
            description += f' switched to agressive mixing scheme (IMIX={last_mixing_scheme})'

        # add mixing scheme
        new_params['IMIX'] = last_mixing_scheme
        self.ctx.last_mixing_scheme = last_mixing_scheme

        if switch_higher_accuracy:
            self.ctx.kkr_higher_accuracy = True
            convergence_settings = self.ctx.convergence_setting_fine
            label += ' use_higher_accuracy'
            description += ' using higher accuracy settings goven in convergence_setting_fine'
        else:
            convergence_settings = self.ctx.convergence_setting_coarse

        # slightly increase temperature if previous calculation was
        # unsuccessful for the second time
        if decrease_mixing_fac and not self.convergence_on_track():
            self.report(
                'INFO: last calculation did not converge and convergence not on track. Try to increase temperature by 50K.'
            )
            convergence_settings['tempr'] += self.ctx.tempr_increase
            label += ' TEMPR_increased'
            description += f' with increased temperature of {self.ctx.tempr_increase}K'

        # add convegence settings
        if self.ctx.loop_count == 1 or self.ctx.last_mixing_scheme == 0:
            new_params['QBOUND'] = self.ctx.threshold_aggressive_mixing
        else:
            new_params['QBOUND'] = self.ctx.convergence_criterion
        new_params['NPOL'] = convergence_settings['npol']
        new_params['NPT1'] = convergence_settings['n1']
        new_params['NPT2'] = convergence_settings['n2']
        new_params['NPT3'] = convergence_settings['n3']
        new_params['TEMPR'] = convergence_settings['tempr']
        new_params['BZDIVIDE'] = convergence_settings['kmesh']

        return new_params, label, description

    def initial_mag(self, new_params, initial_settings):
        """Add settings for initial magnetization"""
        if initial_settings and self.ctx.mag_init:
            if self.ctx.hfield <= 0:
                self.report(
                    f'\nWARNING: magnetization initialization chosen but hfield is zero. '
                    f'Automatically change back to default value (hfield={self._wf_default["hfield"]})\n'
                )
                self.ctx.hfield = self._wf_default['hfield']
            xinipol = self.ctx.xinit
            if xinipol is None:
                # find structure to determine needed length on xinipol
                if 'structure' in self.inputs:
                    struc = self.inputs.structure
                else:
                    struc, voro_parent = VoronoiCalculation.find_parent_structure(self.ctx.last_remote)
                natom = len(get_site_symbols(struc))
                xinipol = np.ones(natom)
            new_params['LINIPOL'] = True
            new_params['HFIELD'] = self.ctx.hfield
            new_params['XINIPOL'] = xinipol
        # turn off initialization after first (successful) iteration
        elif self.ctx.mag_init and self.ctx.mag_init_step_success:
            new_params['LINIPOL'] = False
            new_params['HFIELD'] = 0.0
        elif not self.ctx.mag_init:
            self.report("INFO: mag_init is False. Overwrite 'HFIELD' to '0.0' and 'LINIPOL' to 'False'.")
            # reset mag init to avoid reinitializing
            new_params['HFIELD'] = 0.0
            new_params['LINIPOL'] = False

        # set nspin to 2 if mag_init is used
        if self.ctx.mag_init:
            nspin_in = new_params.get('NSPIN')
            if nspin_in is None:
                nspin_in = 1
            if nspin_in < 2:
                self.report('WARNING: found NSPIN=1 but for maginit needs NPIN=2. Overwrite this automatically')
                new_params['NSPIN'] = 2

        return new_params

    def run_kkr(self):
        """
        submit a KKR calculation
        """
        self.report(f'INFO: setting up kkr calculation step {self.ctx.loop_count}')

        label = f'KKR calculation step {self.ctx.loop_count} (IMIX={self.ctx.last_mixing_scheme})'
        description = f'KKR calculation of step {self.ctx.loop_count}, using mixing scheme {self.ctx.last_mixing_scheme}'
        code = self.inputs.kkr
        remote = self.ctx.last_remote
        params = self.ctx.last_params.get_dict()

        # overwrite some parameters of the KKR calculation by hand before setting mandatory keys
        if 'params_kkr_overwrite' in self.inputs:
            for key, val in self.inputs.params_kkr_overwrite.get_dict().items():
                params[key] = val
                self.report(f'INFO: overwriting KKR parameter: {key} with {val} from params_kkr_overwrite input node')

        options = {
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'resources': self.ctx.resources,
            'queue_name': self.ctx.queue
        }
        if self.ctx.custom_scheduler_commands:
            options['custom_scheduler_commands'] = self.ctx.custom_scheduler_commands

        inputs = get_inputs_kkr(
            code,
            remote,
            options,
            label,
            description,
            parameters=orm.Dict(params),
            serial=(not self.ctx.withmpi),
        )

        # pass nonco angles setting to KkrCalculation
        if 'initial_noco_angles' in self.inputs:
            inputs['initial_noco_angles'] = self.ctx.initial_noco_angles

        # run the KKR calculation
        self.report('INFO: doing calculation')
        kkr_run = self.submit(KkrCalculation, **inputs)

        return ToContext(kkr=kkr_run, last_calc=kkr_run)

    def inspect_kkr(self):
        """Check for convergence and store the results of the last calculation."""
        self.report('INFO: inspecting kkr results step')

        self.report(f'Caching info: {self.ctx.last_calc.get_cache_source()}')

        self.ctx.calcs.append(self.ctx.last_calc)
        self.ctx.kkr_step_success = True

        # check calculation state
        if not self.ctx.last_calc.is_finished_ok:
            self.ctx.kkr_step_success = False
            self.report('ERROR: last calculation not finished correctly')

        self.report(f'INFO: kkr_step_success: {self.ctx.kkr_step_success}')

        # extract convergence info about rms etc. (used to determine convergence behavior)
        try:
            self.report(f'INFO: trying to find output of last_calc: {self.ctx.last_calc}')
            last_calc_output = self.ctx.last_calc.outputs.output_parameters.get_dict()
            found_last_calc_output = True
        except:
            found_last_calc_output = False
        self.report(f'INFO: found_last_calc_output: {found_last_calc_output}')

        # try to extract remote folder
        try:
            self.ctx.last_remote = self.ctx.kkr.outputs.remote_folder
        except:
            self.ctx.last_remote = None
            self.ctx.kkr_step_success = False

        self.report(f'INFO: last_remote: {self.ctx.last_remote}')

        if self.ctx.kkr_step_success and found_last_calc_output:
            # check rms, compare spin and charge values and take bigger one
            rms_charge = last_calc_output['convergence_group']['rms']
            # returning 0 if not found allows to reuse older verisons (e.g. in caching)
            rms_spin = last_calc_output['convergence_group'].get('rms_spin', 0)
            if rms_spin is None:
                rms_spin = 0  # this happens for NSPIN==1
            if rms_charge >= rms_spin:
                rms_max = rms_charge
                use_rms_charge = True
            else:
                rms_max = rms_spin
                use_rms_charge = False
            self.ctx.rms.append(rms_max)
            if use_rms_charge:
                rms_all_iter_last_calc = list(last_calc_output['convergence_group']['rms_all_iterations'])
            else:
                rms_all_iter_last_calc = list(last_calc_output['convergence_group']['rms_spin_all_iterations'])
            # check charge neutrality
            self.ctx.neutr.append(last_calc_output['convergence_group']['charge_neutrality'])
            neutr_all_iter_last_calc = list(last_calc_output['convergence_group']['charge_neutrality_all_iterations'])

            # add lists of last iterations
            self.ctx.last_rms_all = rms_all_iter_last_calc
            self.ctx.last_neutr_all = neutr_all_iter_last_calc
            if self.ctx.kkr_step_success and self.convergence_on_track():
                self.ctx.rms_all_steps += rms_all_iter_last_calc
                self.ctx.neutr_all_steps += neutr_all_iter_last_calc

            # check if calculation converged
            self.ctx.kkr_converged = True
            if rms_max > self.ctx.convergence_criterion:
                self.ctx.kkr_converged = False

        else:
            # if last step did not succeed we know the calculation did not converge
            self.ctx.kkr_converged = False

        self.report(f'INFO: kkr_converged: {self.ctx.kkr_converged}')
        self.report(f'INFO: rms: {self.ctx.rms}')
        self.report(f'INFO: last_rms_all: {self.ctx.last_rms_all}')
        #self.report("INFO: rms_all_steps: {}".format(self.ctx.rms_all_steps))
        self.report(f'INFO: charge_neutrality: {self.ctx.neutr}')
        self.report(f'INFO: last_neutr_all: {self.ctx.last_neutr_all}')
        #self.report("INFO: neutr_all_steps: {}".format(self.ctx.neutr_all_steps))

        # turn off initial magnetization once one step was successful (update_kkr_params) used in
        if self.ctx.mag_init and self.ctx.kkr_step_success:
            self.ctx.mag_init_step_success = True

        # TODO: extract something else (maybe total energy, charge neutrality, magnetisation)?

        # store some statistics used to print table in the end of the report
        tmplist = self.ctx.KKR_steps_stats.get('success', [])
        self.report(f'INFO: append kkr_step_success {tmplist}, {self.ctx.kkr_step_success}')
        tmplist.append(self.ctx.kkr_step_success)
        self.ctx.KKR_steps_stats['success'] = tmplist
        try:
            isteps = self.ctx.last_calc.outputs.output_parameters.get_dict(
            )['convergence_group']['number_of_iterations']
        except (AttributeError, NotExistentAttributeError):
            self.ctx.warnings.append('cound not set isteps in KKR_steps_stats dict')
            isteps = -1

        try:
            first_rms = self.ctx.last_rms_all[0]
            last_rms = self.ctx.last_rms_all[-1]
        except:
            self.ctx.warnings.append('cound not set first_rms, last_rms in KKR_steps_stats dict')
            first_rms = -1
            last_rms = -1

        try:
            first_neutr = self.ctx.last_neutr_all[0]
            last_neutr = self.ctx.last_neutr_all[-1]
        except:
            self.ctx.warnings.append('cound not set first_neutr, last_neutr in KKR_steps_stats dict')
            first_neutr = -999
            last_neutr = -999

        if self.ctx.last_mixing_scheme == 0:
            mixfac = self.ctx.strmix
        else:
            mixfac = self.ctx.brymix

        if self.ctx.kkr_higher_accuracy:
            qbound = self.ctx.convergence_criterion
        else:
            qbound = self.ctx.threshold_switch_high_accuracy

        # store results in KKR_steps_stats dict
        for key, val in {
            'isteps': isteps,
            'imix': self.ctx.last_mixing_scheme,
            'mixfac': mixfac,
            'qbound': qbound,
            'high_sett': self.ctx.kkr_higher_accuracy,
            'first_rms': first_rms,
            'last_rms': last_rms,
            'first_neutr': first_neutr,
            'last_neutr': last_neutr,
            'pk': self.ctx.last_calc.pk,
            'uuid': self.ctx.last_calc.uuid
        }.items():
            tmplist = self.ctx.KKR_steps_stats.get(key, [])
            tmplist.append(val)
            self.ctx.KKR_steps_stats[key] = tmplist

        # update noco angles
        if 'initial_noco_angles' in self.inputs and (self.ctx.kkr_step_success and found_last_calc_output):
            self._get_new_noco_angles()

        self.report('INFO: done inspecting kkr results step')

    def convergence_on_track(self):
        """Check if convergence behavior of the last calculation is on track.

        (i.e. going down)
        """
        on_track = True
        threshold = 5.  # used to check condition if at least one of charnge_neutrality, rms-error goes down fast enough
        eps_neutr = 1e-4
        eps_rms = 1e-6

        # first check if previous calculation was stopped due to reaching the QBOUND limit
        try:
            calc_reached_qbound = self.ctx.last_calc.outputs.output_parameters.get_dict(
            )['convergence_group']['calculation_converged']
        except:  # captures error when last_calc dies not have an output node
            calc_reached_qbound = False

        if self.ctx.kkr_step_success and not calc_reached_qbound:
            first_rms = self.ctx.last_rms_all[0]
            first_neutr = abs(self.ctx.last_neutr_all[0])
            last_rms = self.ctx.last_rms_all[-1]
            last_neutr = abs(self.ctx.last_neutr_all[-1])
            # use this trick to avoid division by zero
            if first_neutr == 0:
                first_neutr = 10**-16
            if first_rms == 0:
                first_rms = 10**-16
            r, n = last_rms / first_rms, last_neutr / first_neutr

            # deal with small differences
            if abs(first_neutr - last_neutr) < eps_neutr:
                n = 0
            if abs(first_rms - last_rms) > eps_rms:
                r = 0

            self.report(
                f'INFO convergence check: first/last rms {first_rms},'
                f' {last_rms}; first/last neutrality {first_neutr}, {last_neutr}'
            )
            if r < 1 and n < 1:
                self.report('INFO convergence check: both rms and neutrality go down')
                on_track = True
            elif n > threshold or r > threshold:
                self.report('INFO convergence check: rms or neutrality goes up too fast, convergence is not expected')
                on_track = False
            elif n * r < 1:
                self.report('INFO convergence check: either rms goes up and neutrality goes down or vice versa')
                self.report('INFO convergence check: but product goes down fast enough')
                on_track = True
            elif len(self.ctx.last_rms_all) == 1:
                self.report('INFO convergence check: already converged after single iteration')
                on_track = True
            else:
                self.report(
                    'INFO convergence check: rms or neutrality do not shrink fast enough, convergence is not expected'
                )
                on_track = False

        elif calc_reached_qbound:
            self.report('INFO convergence check: calculation reached QBOUND')
            on_track = True
        else:
            self.report('INFO convergence check: calculation unsuccessful')
            on_track = False

        self.report(f'INFO convergence check result: {on_track}')

        return on_track

    def return_results(self):
        """Return the results of the calculations.

        This should run through and produce output nodes even if everything failed,
        therefore it only uses results from context.
        """

        self.report('INFO: entering return_results')

        # try/except to capture as mnuch as possible
        # (everything that is there even when workflow exits unsuccessfully)
        # capture pk and uuids of last calc, params and remote
        try:
            last_calc_uuid = self.ctx.last_calc.uuid
            last_calc_pk = self.ctx.last_calc.pk
            last_params_uuid = self.ctx.last_params.uuid
            last_params_pk = self.ctx.last_params.pk
            last_remote_uuid = self.ctx.last_remote.uuid
            last_remote_pk = self.ctx.last_remote.pk
        except:
            last_calc_uuid = None
            last_calc_pk = None
            last_params_uuid = None
            last_params_pk = None
            last_remote_uuid = None
            last_remote_pk = None

        all_pks = []
        for calc in self.ctx.calcs:
            try:
                all_pks.append(calc.pk)
            except:
                self.ctx.warnings.append(f'cound not get pk of calc {calc}')

        # capture links to last parameter, calcualtion and output
        try:
            last_calc_out = self.ctx.kkr.outputs.output_parameters
            last_calc_out_dict = last_calc_out.get_dict()
            self.report('Found last_calc_out')
            last_RemoteData = self.ctx.last_remote
            self.report('Found last_remote')
            last_InputParameters = self.ctx.last_params
            self.report('Found last_params')
        except:
            self.report('Error in finding last_calc_out etc.')
            last_InputParameters = None
            last_RemoteData = None
            last_calc_out = None
            last_calc_out_dict = {}

        # capture convergence info
        try:
            last_rms = self.ctx.rms[-1]
            last_neutr = self.ctx.neutr[-1]
        except:
            last_rms = None
            last_neutr = None

        # capture result of vorostart sub-workflow
        try:
            results_vorostart = self.ctx.voronoi.outputs.results_vorostart_wc
        except:
            results_vorostart = None
        try:
            starting_dosdata_interpol = self.ctx.voronoi.outputs.last_doscal_dosdata_interpol
        except:
            starting_dosdata_interpol = None

        try:
            final_dosdata_interpol = self.ctx.doscal.outputs.dos_data_interpol
        except:
            final_dosdata_interpol = None

        # now collect results saved in results node of workflow
        self.report('INFO: collect outputnode_dict')
        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['material'] = self.ctx.formula
        outputnode_dict['loop_count'] = self.ctx.loop_count
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['last_params_nodeinfo'] = {'uuid': last_params_uuid, 'pk': last_params_pk}
        outputnode_dict['last_remote_nodeinfo'] = {'uuid': last_remote_uuid, 'pk': last_remote_pk}
        outputnode_dict['last_calc_nodeinfo'] = {'uuid': last_calc_uuid, 'pk': last_calc_pk}
        outputnode_dict['pks_all_calcs'] = all_pks
        outputnode_dict['errors'] = self.ctx.errors
        outputnode_dict['convergence_value'] = last_rms
        outputnode_dict['convergence_values_all_steps'] = np.array(self.ctx.rms_all_steps)
        outputnode_dict['convergence_values_last_step'] = np.array(self.ctx.last_rms_all)
        outputnode_dict['charge_neutrality'] = last_neutr
        outputnode_dict['charge_neutrality_all_steps'] = np.array(self.ctx.neutr_all_steps)
        outputnode_dict['charge_neutrality_last_step'] = np.array(self.ctx.last_neutr_all)
        outputnode_dict['dos_check_ok'] = self.ctx.dos_ok
        outputnode_dict['convergence_reached'] = self.ctx.kkr_converged
        outputnode_dict['voronoi_step_success'] = self.ctx.voro_step_success
        outputnode_dict['kkr_step_success'] = self.ctx.kkr_step_success
        outputnode_dict['used_higher_accuracy'] = self.ctx.kkr_higher_accuracy

        # report the status
        if self.ctx.successful:
            self.report(
                'STATUS: Done, the convergence criteria are reached.\n'
                f'INFO: The charge density of the KKR calculation pk= {last_calc_pk} '
                f'converged after {self.ctx.loop_count} KKR runs and '
                f'{self.ctx.loop_count} iterations to {last_rms} \n'
            )
        else:  # Termination ok, but not converged yet...
            if self.ctx.abort:  # some error occured, donot use the output.
                self.report('STATUS/ERROR: I abort, see logs and '
                            'erros/warning/hints in output_kkr_scf_wc_para')
            else:
                self.report(
                    'STATUS/WARNING: Done, the maximum number of runs '
                    'was reached or something failed.\n INFO: The '
                    'charge density of the KKR calculation pk= '
                    f'after {self.ctx.loop_count} KKR runs and '
                    f"{sum(self.ctx.KKR_steps_stats.get('isteps')),} "
                    f"iterations is {last_rms} 'me/bohr^3'\n"
                )

        # create results  node
        # : {}".format(outputnode_dict))
        self.report('INFO: create results node')
        outputnode_t = orm.Dict(dict=outputnode_dict)
        outputnode_t.label = 'kkr_scf_wc_results'
        outputnode_t.description = (
            'Contains results of workflow'
            ' (e.g. workflow version number, info about success of wf,'
            ' lis tof warnings that occured during execution, ...)'
        )

        # collect nodes in outputs dictionary
        out_nodes = {'outpara': outputnode_t}
        if last_calc_out is not None:
            out_nodes['last_calc_out'] = last_calc_out
        if last_RemoteData is not None:
            out_nodes['last_RemoteData'] = last_RemoteData
        if last_InputParameters is not None:
            out_nodes['last_InputParameters'] = last_InputParameters
        if final_dosdata_interpol is not None:
            out_nodes['final_dosdata_interpol'] = final_dosdata_interpol
        if starting_dosdata_interpol is not None:
            out_nodes['starting_dosdata_interpol'] = starting_dosdata_interpol
        if results_vorostart is not None:
            out_nodes['results_vorostart'] = results_vorostart
        if 'initial_noco_angles' in self.inputs and not all(self.ctx.initial_noco_angles['fix_dir']):
            # was updated in inspect_kkr to last noco angles output
            out_nodes['last_noco_angles'] = self.ctx.initial_noco_angles

        # call helper function to create output nodes in correct AiiDA graph structure
        outdict = create_scf_result_node(**out_nodes)

        for link_name, node in outdict.items():
            #self.report("INFO: storing node {} {} with linkname {}".format(type(node), node, link_name))
            self.out(link_name, node)

        # print results table for overview
        # table layout:
        message = 'INFO: overview of the result:\n\n'
        message += '|------|---------|--------|------|--------|-------------------|-----------------|-----------------|-------------\n'
        message += '| irun | success | isteps | imix | mixfac | accuracy settings |       rms       | abs(neutrality) | pk and uuid \n'
        message += '|      |         |        |      |        | qbound  | higher? | first  |  last  | first  |  last  |             \n'
        message += '|------|---------|--------|------|--------|---------|---------|--------|--------|--------|--------|-------------\n'
        # | %6i  | %9s     | %8i    | %6i  | %.2e   | %.3e    | %9s     | %.2e   |  %.2e  |  %.2e  |  %.2e  |
        KKR_steps_stats = self.ctx.KKR_steps_stats
        for irun in range(len(KKR_steps_stats.get('success'))):
            KKR_steps_stats.get('first_neutr')[irun] = abs(KKR_steps_stats.get('first_neutr')[irun])
            KKR_steps_stats.get('last_neutr')[irun] = abs(KKR_steps_stats.get('last_neutr')[irun])
            message += (
                '|{:6d}|{:9s}|{:8d}|{:6d}|{:.2e}|{:.3e}|{:9s}|{:.2e}|{:.2e}|{:.2e}|{:.2e}|'.format(
                    irun + 1, str(KKR_steps_stats.get('success')[irun]),
                    KKR_steps_stats.get('isteps')[irun],
                    KKR_steps_stats.get('imix')[irun],
                    KKR_steps_stats.get('mixfac')[irun],
                    KKR_steps_stats.get('qbound')[irun], str(KKR_steps_stats.get('high_sett')[irun]),
                    KKR_steps_stats.get('first_rms')[irun],
                    KKR_steps_stats.get('last_rms')[irun],
                    KKR_steps_stats.get('first_neutr')[irun],
                    KKR_steps_stats.get('last_neutr')[irun]
                )
            )
            message += f" {KKR_steps_stats.get('pk')[irun]} | {KKR_steps_stats.get('uuid')[irun]}\n"
            """
            message += "#|{}|{}|{}|{}|{}|{}|{}|{}|{}|{}|{}|\n".format(irun+1,
                          KKR_steps_stats.get('success')[irun], KKR_steps_stats.get('isteps')[irun],
                          KKR_steps_stats.get('imix')[irun], KKR_steps_stats.get('mixfac')[irun],
                          KKR_steps_stats.get('qbound')[irun], KKR_steps_stats.get('high_sett')[irun],
                          KKR_steps_stats.get('first_rms')[irun], KKR_steps_stats.get('last_rms')[irun],
                          KKR_steps_stats.get('first_neutr')[irun], KKR_steps_stats.get('last_neutr')[irun])
            """
        self.report(message)

        self.report('\nINFO: done with kkr_scf workflow!\n')

    def check_input_params(self, params, is_voronoi=False):
        """Check the input parameter consistency and aborts wf if check fails."""
        if params is None:
            return self.exit_codes.ERROR_NO_CALC_PARAMETERS_GIVEN  # pylint: disable=no-member
        input_dict = params.get_dict()
        if is_voronoi:
            para_check = kkrparams(params_type='voronoi')
        else:
            para_check = kkrparams()

        # step 1 try to fill keywords
        try:
            for key, val in input_dict.items():
                para_check.set_value(key, val, silent=True)
        except:
            return self.exit_codes.ERROR_CALC_PARAMTERS_INCONSISTENT  # pylint: disable=no-member

        # step 2: check if all mandatory keys are there
        missing_list = para_check.get_missing_keys(use_aiida=True)
        if missing_list != []:
            all_defaults = True
            for key in missing_list:
                if key not in kkrparams.get_KKRcalc_parameter_defaults()[0]:
                    all_defaults = False
            if not all_defaults:
                self.report(
                    f'ERROR: calc_parameters given are not consistent! '
                    f'Missing mandatory keys: {missing_list}'
                )
                return self.exit_codes.ERROR_CALC_PARAMETERS_INCOMPLETE  # pylint: disable=no-member
        return None

    def get_dos(self):
        """Call the dos sub workflow pass the input and submit the calculation."""
        if self.ctx.check_dos:
            self.report('INFO: Doing final DOS calculation')

            # fix emin/emax to include emin, ef of scf contour
            # remember: efermi, emin and emax are in internal units (Ry) but delta_e is in eV!
            emin = self.ctx.dos_params['emin']  # from dos params
            # from contour (sets limit of dos emin!)
            emin_cont = self.ctx.last_calc.outputs.output_parameters.get_dict().get('energy_contour_group').get('emin')
            if emin_cont - self.ctx.delta_e * eV2Ry < emin:
                self.ctx.dos_params['emin'] = (emin_cont - self.ctx.delta_e * eV2Ry)
                self.report(
                    f'INFO: emin ({emin_cont} Ry) - '
                    f'delta_e ({self.ctx.delta_e * eV2Ry} Ry) smaller than '
                    f'emin ({emin} Ry) of voronoi output. '
                    f'Setting automatically to {emin_cont - self.ctx.delta_e * eV2Ry}Ry'
                )
            with self.ctx.last_calc.outputs.retrieved.open('out_potential') as _file:
                self.ctx.efermi = get_ef_from_potfile(_file)
            emax = self.ctx.dos_params['emax']
            if emax < self.ctx.efermi + self.ctx.delta_e * eV2Ry:
                self.ctx.dos_params['emax'] = self.ctx.efermi + self.ctx.delta_e * eV2Ry
                self.report(
                    f'INFO: self.ctx.efermi ({self.ctx.efermi} Ry) '
                    f'+ delta_e ({self.ctx.delta_e * eV2Ry} Ry) larger than '
                    f'emax ({emax} Ry). Setting automatically to '
                    f'{self.ctx.efermi + self.ctx.delta_e * eV2Ry}Ry'
                )

            wfdospara_node = orm.Dict(dict=self.ctx.dos_params)
            wfdospara_node.label = 'DOS params'
            wfdospara_node.description = 'DOS parameter set for final DOS calculation of kkr_scf_wc'

            code = self.inputs.kkr
            remote = self.ctx.last_calc.outputs.remote_folder
            wf_label = ' final DOS calculation'
            wf_desc = ' subworkflow of a DOS calculation'

            builder = kkr_dos_wc.get_builder()
            builder.metadata.description = wf_desc  # pylint: disable=no-member
            builder.metadata.label = wf_label  # pylint: disable=no-member
            builder.kkr = code
            builder.wf_parameters = wfdospara_node
            builder.options = self.ctx.options_params_dict
            builder.remote_data = remote

            future = self.submit(builder)

            return ToContext(doscal=future)

    def check_dos(self):
        """Checks if dos of final potential is ok."""
        # initialize dos_ok variable
        self.ctx.dos_ok = True

        # first check if dos should be checked or if the test is skipped
        if not self.ctx.check_dos:
            self.report('INFO: skipping DOS check')
            return

        self.report('INFO: checking DOS for consistency (EMIN position, negative DOS, etc.)')

        # check parser output
        doscal = self.ctx.doscal
        try:
            dos_outdict = doscal.outputs.results_wf.get_dict()
            if not dos_outdict['successful']:
                self.report('ERROR: DOS workflow unsuccessful')
                self.ctx.dos_ok = False
                return self.exit_codes.ERROR_DOS_RUN_UNSUCCESSFUL  # pylint: disable=no-member

            if dos_outdict['list_of_errors'] != []:
                self.report(f'ERROR: DOS wf output contains errors: {dos_outdict["list_of_errors"]}')
                self.ctx.dos_ok = False
                return self.exit_codes.ERROR_DOS_RUN_UNSUCCESSFUL  # pylint: disable=no-member
        except AttributeError:
            self.ctx.dos_ok = False
            return self.exit_codes.ERROR_DOS_RUN_UNSUCCESSFUL  # pylint: disable=no-member

        # check for negative DOS
        try:
            dosdata = doscal.outputs.dos_data
            natom = self.ctx.last_calc.outputs.output_parameters.get_dict()['number_of_atoms_in_unit_cell']
            nspin = dos_outdict['nspin']

            ener = dosdata.get_x()[1]
            totdos = dosdata.get_y()[0][1]

            if len(ener) != nspin * natom:
                self.report(
                    f'ERROR: DOS output shape does not fit nspin, natom '
                    f'information: len(energies)={len(ener)}, natom={natom}, nspin={nspin}'
                )
                self.ctx.doscheck_ok = False
                return self.exit_codes.ERROR_DOS_RUN_UNSUCCESSFUL  # pylint: disable=no-member

            # deal with snpin==1 or 2 cases and check negtive DOS
            for iatom in range(natom // nspin):
                for ispin in range(nspin):
                    x, y = ener[iatom * nspin + ispin], totdos[iatom * nspin + ispin]
                    if nspin == 2 and ispin == 0:
                        y = -y
                    if y.min() < 0:
                        self.report(
                            f'INFO: negative DOS value found in '
                            f'(atom, spin)=({iatom},{ispin}) at iteration '
                            f'{self.ctx.loop_count}: {y.min()}'
                        )
                        self.ctx.dos_ok = False

            # check starting EMIN
            dosdata_interpol = doscal.outputs.dos_data_interpol

            ener = dosdata_interpol.get_x()[1]  # shape= natom*nspin, nept
            totdos = dosdata_interpol.get_y()[0][1]  # shape= natom*nspin, nept
            Ry2eV = get_Ry2eV()

            for iatom in range(natom // nspin):
                for ispin in range(nspin):
                    x, y = ener[iatom * nspin + ispin], totdos[iatom * nspin + ispin]
                    xrel = abs(x - (self.ctx.dos_params_dict['emin'] - self.ctx.efermi) * Ry2eV)
                    mask_emin = np.where(xrel == xrel.min())
                    ymin = abs(y[mask_emin])
                    if ymin > self.ctx.threshold_dos_zero:
                        self.report(f'INFO: DOS at emin not zero! {ymin}>{self.ctx.threshold_dos_zero}')
                        self.ctx.dos_ok = False
        except AttributeError:
            self.ctx.dos_ok = False

    def _get_new_noco_angles(self):
        """Extract nonco angles from output of calculation.

        If fix_dir is True we skip this and leave the initial angles unchanged
        Here we update self.ctx.initial_noco_angles with the new values
        """
        # first check if we need to update the angles
        if all(self.ctx.initial_noco_angles['fix_dir']):
            return

        # extract output angles from last calculation and update initial_noco_angles in context
        self.ctx.initial_noco_angles = extract_noco_angles(
            fix_dir_threshold=orm.Float(self.ctx.fix_dir_threshold),
            old_noco_angles=self.ctx.initial_noco_angles,
            last_retrieved=self.ctx.last_calc.outputs.retrieved
        )
        if self.ctx.initial_noco_angles == {}:
            self.ctx.initial_noco_angles = self.ctx.initial_noco_angles


@workfunction
def create_scf_result_node(**kwargs):
    """This is a pseudo wf, to create the right graph structure of AiiDA.

    This workfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """

    has_last_outpara = False
    has_last_calc_out_dict = False
    has_last_RemoteData = False
    has_vorostart_output = False
    has_starting_dos = False
    has_final_dos = False
    has_last_InputParameters = False

    for key, val in kwargs.items():
        if key == 'outpara':  # should always be there
            outpara = val
            has_last_outpara = True
        elif key == 'last_calc_out':
            has_last_calc_out_dict = True
            last_calc_out_dict = val
        elif key == 'last_RemoteData':
            last_RemoteData_dict = val
            has_last_RemoteData = True
        elif key == 'last_InputParameters':
            last_InputParameters_dict = val
            has_last_InputParameters = True
        elif key == 'results_vorostart':
            has_vorostart_output = True
            vorostart_output_dict = val
        elif key == 'starting_dosdata_interpol':
            has_starting_dos = True
            start_dosdata_interpol_dict = val
        elif key == 'final_dosdata_interpol':
            has_final_dos = True
            final_dosdata_interpol_dict = val

    outdict = {}

    if has_last_outpara:
        outputnode = outpara  # pylint: disable=possibly-used-before-assignment
        outputnode.label = 'workflow_Results'
        outputnode.description = ('Contains self-consistency results and '
                                  'information of an kkr_scf_wc run.')
        outdict['output_kkr_scf_wc_ParameterResults'] = outputnode

    if has_last_calc_out_dict:
        outputnode = last_calc_out_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'last_calc_out'
        outputnode.description = (
            'Contains the Results Parameter node from the output '
            'of the last calculation done in the workflow.'
        )
        outdict['last_calc_out'] = outputnode

    if has_last_RemoteData:
        outputnode = last_RemoteData_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'last_RemoteData'
        outputnode.description = (
            'Contains a link to the latest remote data node '
            'where the output of the calculation can be accessed.'
        )
        outdict['last_RemoteData'] = outputnode

    if has_last_InputParameters:
        outputnode = last_InputParameters_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'last_InputParameters'
        outputnode.description = (
            'Contains the latest parameter data node '
            'where the input of the last calculation can be found.'
        )
        outdict['last_InputParameters'] = outputnode

    if has_vorostart_output:
        outputnode = vorostart_output_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'results_vorostart'
        outputnode.description = (
            'Contains the results parameter data node '
            'of the vorostart sub-workflow (sets up starting portentials).'
        )
        outdict['results_vorostart'] = outputnode

    if has_starting_dos:
        outputnode = start_dosdata_interpol_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'starting_dosdata_interpol'
        outputnode.description = ('Contains the interpolated DOS data note, computed '
                                  'from the starting portential.')
        outdict['starting_dosdata_interpol'] = outputnode

    if has_final_dos:
        outputnode = final_dosdata_interpol_dict  # pylint: disable=used-before-assignment
        outputnode.label = 'final_dosdata_interpol'
        outputnode.description = ('Contains the interpolated DOS data note, computed '
                                  'from the converged potential.')
        outdict['final_dosdata_interpol'] = outputnode

    return outdict


def get_site_symbols(structure):
    """
    extract the site number taking into account a possible CPA structure
    """
    sites = structure.sites
    sitelist = []
    for isite, site in enumerate(sites):
        sitekind = structure.get_kind(site.kind_name)
        for ikind in range(len(sitekind.symbols)):
            site_symbol = sitekind.symbols[ikind]
            sitelist.append([isite, site_symbol])

    return sitelist



"""
Tools for Jij calculations and parsing
"""

# parsing jij output

import numpy as np
from masci_tools.io.common_functions import search_string, get_aBohr2Ang, get_Ry2eV
from aiida.orm import ArrayData, StructureData, Dict, load_node, Bool
from aiida.engine import calcfunction, submit
from aiida_kkr.calculations import KkrCalculation
from aiida_kkr.tools import find_parent_structure


def get_sites(structure):
    """Get all sites also for a CPA structure"""
    sites = []  # for CPA
    for site in structure.sites:
        sitekind = structure.get_kind(site.kind_name)
        for ikind in range(len(sitekind.symbols)):
            sites.append(site)
    return sites


def get_jijs_shells(jij_calc, verbose=False):
    """read the jij.atom files from the retrieved and determine if the file has DMI or not"""

    # read the jij.atom files
    jijs_shells_all = []
    for jijfile in [i for i in jij_calc.outputs.retrieved.list_object_names() if 'Jij' in i]:
        i_index = int(jijfile.split('.atom')[1])
        if verbose:
            print('load jij.atom', i_index)
        with jij_calc.outputs.retrieved.open(jijfile) as f:
            try:
                tmp = np.loadtxt(f)
            except:
                # this means we could not read the file which happens, for example,
                # in case of CPA where "&" characters are inserted to mark the end of block
                f.seek(0)
                txt = f.readlines()
                txt = [line for line in txt if '&' not in line]
                tmp = np.loadtxt(txt)
        # sort by radius
        jij_atom = tmp[tmp[:, 0].argsort()]
        if len(jij_atom[0]) > 4:
            # add i index, needed for mapping to spirit data
            # only needed for DMI mode
            jij_atom[:, -1] = i_index
        if verbose:
            print('shape jij.atom', jij_atom.shape)
        jijs_shells_all += jij_atom.tolist()

    if verbose:
        print('jijs_shells.shape', len(jijs_shells_all))

    if len(jijs_shells_all) == 0:
        raise ValueError('Did not extract any Jij shells')

    # format of jijs_shells differs in old and new solver:
    if len(jijs_shells_all[0]) == 4:
        # - isotropic exchange constants only (old solver). The columns refer to:
        #   * [0] |Rij| in units of the lattice constant
        #   * [1] Jij in Ryd
        #   * [2] shell-number
        #   * [3] atom type of atom j
        #   * [4] atom type of atom i
        dmimode = False
    else:
        # - full exchange tensor (new solver). The columns refer to:
        #   * [0] ∣Rij∣ in units of the lattice constant
        #   * [1] Jij (isotropic part) in Ryd
        #   * [2] Dij (anti-symmetric DMI part) in Ryd
        #   * [3] Sij (diagonal traceless part) in Ryd
        #   * [4] Aij (off-diagonal symmetric part) in Ryd
        #   * [5-7] Rj− Ri (3 component vector)
        #   * [8] atom type of atom j
        #   * [9] atom type of atom i
        dmimode = True

    return jijs_shells_all, dmimode


def expand_jijs_iso(jij_calc, jijs_shells_all, shells, alat):
    """expand jijs from shells to all pairs which is easier to use in spirit
    This is the roune that deals with isotropic interactions (i.e. using the old solver)"""

    # extract basis vectors ot the structure, needed to convert from cartesian coordinates (x,y,z)
    # to relative coordinates (da, db, dc) such that (x, y, z)^T = da*a + db*b + dc*c
    # with a,b,c being the three unit vectors
    structure = find_parent_structure(jij_calc)
    cell = np.array(structure.cell)
    cell_positions = np.array([i.position for i in structure.sites])

    # transformation matrix from absolute to realtive coordinates
    U2rel = np.linalg.inv(cell.transpose())

    # expand shells data to all structure that spirit can understand (i.e. take all pairs)
    jijs_expanded, positions_expanded = [], []
    for jshell in jijs_shells_all:
        for s in shells[int(jshell[3] - 1)]:
            # i and j indices in the unit cell
            iatom, jatom = int(s[0] - 1), int(s[1] - 1)

            # absolute distance
            # Rx, Ry, Rz = alat*(s[7]-s[4]), alat*(s[8]-s[5]), alat*(s[9]-s[6])
            Rx, Ry, Rz = alat * s[7], alat * s[8], alat * s[9]

            # substract sub-lattice distance to find correct unit cell scalings
            R = np.array([Rx, Ry, Rz])
            R -= cell_positions[jatom]
            # R -= (cell_positions[jatom] - cell_positions[i])
            x, y, z = R

            # calculate da, db, dc ( i.e. the mutiplicities of the unit vectors)
            da, db, dc = np.array(np.round(np.dot(U2rel, [x, y, z])), dtype=int)

            #                      i,      j,   da, db, dc, Jij (meV)
            jijs_expanded.append([iatom, jatom, da, db, dc, jshell[1] * get_Ry2eV() * 1000])

            positions_expanded.append(np.array([Rx, Ry, Rz]) - cell_positions[iatom])

    # convert to numpy array
    jijs_expanded = np.array(jijs_expanded)
    positions_expanded = np.array(positions_expanded)

    return jijs_expanded, positions_expanded


def expand_jijs_dmi(jijs_shells_x, jijs_shells_y, jijs_shells_z, shells, cell, sites, alat, verbose=False):
    """Bring output Jijs of the new solver into the right form
    (KKR output already is expanded in shells but it needs to be brought into the right form
    by combining the x,y,z calculations and the Rvec information)
    This will give the full Jij tensor and prodiuce the columns that spirit can understand
    """

    # conversion factor from Ry to meV
    Ry2meV = 1000 * get_Ry2eV()

    # transformation matrix from absolute to realtive coordinates
    U2rel = np.linalg.inv(cell.transpose())

    jijs_expanded, positions_expanded = [], []
    for ipos, jz in enumerate(jijs_shells_z):
        # also load the information from m||x and m||y
        jx = jijs_shells_x[ipos]
        jy = jijs_shells_y[ipos]

        if verbose:
            print(ipos, jx, jy, jz)

        # get i and j indices of the atoms
        # -1 to convert from fortran output to python indexing (starting at 0)
        iatom = int(jz[-1]) - 1
        jatom = int(jz[-2]) - 1

        # Rvec = pos_i - pos_j + R_ij[j], with R_ij[j] given in the Jij.atom files (in alat units)
        dx = alat * jz[5]
        dy = alat * jz[6]
        dz = alat * jz[7]
        x = sites[iatom].position[0] - sites[jatom].position[0] + dx
        y = sites[iatom].position[1] - sites[jatom].position[1] + dy
        z = sites[iatom].position[2] - sites[jatom].position[2] + dz

        # calculate da, db, dc ( i.e. the mutiplicities of the unit vectors)
        da, db, dc = np.array(np.round(np.dot(U2rel, [x, y, z])), dtype=int)

        if verbose:
            if ipos == 0:
                print(cell)
            print(iatom, jatom, dx, dy, dz, x, y, z, da, db, dc)

        #collect full Jij tensor
        # 1: J, 2: D, 3: S, 4: A
        Jxx = (jz[1] + jz[3] + jy[1] + jy[3]) / 2. * Ry2meV
        Jxy = jz[4] * Ry2meV
        Jxz = jy[4] * Ry2meV
        Jyx = jz[4] * Ry2meV
        Jyy = (jz[1] - jz[3] + jx[1] + jx[3]) / 2. * Ry2meV
        Jyz = jx[4] * Ry2meV
        Jzx = jy[4] * Ry2meV
        Jzy = jx[4] * Ry2meV
        Jzz = (jx[1] - jx[3] + jy[1] - jy[3]) / 2. * Ry2meV
        # calculate DMI vector
        Dx = jx[2] * Ry2meV
        # for y component no minus sign: ERROR in documentation!
        #Dy = -jy[2] * Ry2meV
        Dy = jy[2] * Ry2meV
        Dz = jz[2] * Ry2meV

        # collect data in big arrays
        #                       i,     j,   da, db, dc,   Jij (meV), Dij vector(meV), full Jij tensor (meV)
        jijs_expanded.append([
            iatom, jatom, da, db, dc, (Jxx + Jyy + Jzz) / 3., Dx, Dy, Dz, Jxx, Jxy, Jxz, Jyx, Jyy, Jyz, Jzx, Jzy, Jzz
        ])
        positions_expanded.append([dx, dy, dz])

    # convert collected data to numpy array
    jijs_expanded = np.array(jijs_expanded)
    positions_expanded = np.array(positions_expanded)

    return jijs_expanded, positions_expanded


def get_jij_structure(structure, jijs_expanded, jij_calc):
    """make auxiliary structure that has only the ij sites from the Jij calculation"""
    struc_jij_sites = StructureData(cell=structure.cell)
    struc_jij_sites.pbc = structure.pbc
    all_sites_jij = set(list(jijs_expanded[:, 0]) + list(jijs_expanded[:, 1]))
    isite, icount, mappings = -1, 0, []  # for mapping to the sites of the reduced structure
    mappings_back = {}
    for site in structure.sites:
        sitekind = structure.get_kind(site.kind_name)
        for ikind, symbol in enumerate(sitekind.symbols):
            isite += 1
            # take only structues for which Jij couplings are extracted
            if isite in all_sites_jij:
                mappings.append([icount, isite])
                mappings_back[isite] = icount
                icount += 1
                if ikind == 0:
                    struc_jij_sites.append_atom(
                        position=site.position, symbols=sitekind.symbols, weights=sitekind.weights
                    )

    # add spin moments extra (used by spirit plugin)
    calc_conv = jij_calc.inputs.parent_folder.get_incoming(node_class=KkrCalculation).first().node
    mu_s_all = calc_conv.outputs.output_parameters['magnetism_group']['spin_moment_per_atom']
    mu_s = []
    for ij in mappings:
        mu_s.append(mu_s_all[ij[1]])
    struc_jij_sites.set_extra('spin_moments', mu_s)

    return struc_jij_sites, mappings_back, mu_s


def get_shells(jij_calc, verbose=False):
    # read the shells information that is needed to map to the complete list of pairs
    with jij_calc.outputs.retrieved.open('shells.dat') as f:
        txt = f.readlines()
    nshell = int(txt.pop(0).split()[0])
    naez = len(find_parent_structure(jij_calc).sites)
    ioffset = 0
    shells = []
    for ishell in range(nshell):
        nat = int(txt[ioffset].split()[1])
        shell = []
        for iline in range(nat):
            shell.append(txt[ioffset + 1 + iline].split())
        shells.append(np.array(shell, dtype=float))
        ioffset += 1 + nat
    if verbose:
        print('found shells:', shells)
    return shells


@calcfunction
def parse_jij_calc_data(
    jij_calc_retrieved, jij_calc_x_retrieved=None, jij_calc_y_retrieved=None, verbose=lambda: Bool(False)
):
    """
    Parse the output of a Jij calculation from the retreived folder of a KkrCalculation

    :params jij_calc_retrieved: retrieved folder output of a KkrCalculation which ran with the Jij inputs (m||z is assumed)
    :params jij_calc_x_retrieved: like jij_calc_retrieved but for m||x (only needed for new solver)
    :params jij_calc_y_retrieved: like jij_calc_retrieved but for m||y (only needed for new solver)
    :params verbose: True/False can be used to print debugging output

    :returns:
      {'jij_data': jij_data, 'structure_jij_sites': struc_jij_sites}
      where
      * jij_data is and ArrayData that contains the expanded Jij's (see 'array_descriptions' extra for more details)
      * struc_jij_sites is the reduced structure that contains onlt the atoms which have Jij couplings
        (comes from the input of the KkrCalculation). The mappings to the original structure (and their i,j indices) is given as an extra
    """
    verbose = verbose.value

    # extract kkrCalculation from retreived child
    jij_calc = jij_calc_retrieved.get_incoming(node_class=KkrCalculation).first().node
    if verbose:
        print('jij(z) calculation:', jij_calc.uuid)

    # extract basis vectors ot the structure, needed to convert from cartesian coordinates (x,y,z)
    # to relative coordinates (da, db, dc) such that (x, y, z)^T = da*a + db*b + dc*c
    # with a,b,c being the three unit vectors
    structure = find_parent_structure(jij_calc)
    cell = np.array(structure.cell)
    natyp = len(get_sites(structure))
    if verbose:
        print('found structure with natyp:', natyp)

    # in KKR everything is scaled with the lattice constant
    alat = jij_calc.outputs.output_parameters['alat_internal'] * get_aBohr2Ang()

    # read jij.atom files
    jijs_shells_z, dmimode = get_jijs_shells(jij_calc, verbose)
    if verbose:
        print('shape jij_shells_z:', len(jijs_shells_z), dmimode)

    # read jij.atom files if a calculation in x and y are given in the input
    dmimode_x, dmimode_y = False, False
    if dmimode and jij_calc_x_retrieved is not None:
        jij_calc_x = jij_calc_x_retrieved.get_incoming(node_class=KkrCalculation).first().node
        if verbose:
            print('jij(x) calculation:', jij_calc_x.uuid)
        jijs_shells_x, dmimode_x = get_jijs_shells(jij_calc_x)
        if verbose:
            print('shape jij_shells_x:', len(jijs_shells_x), dmimode_x)
        if not dmimode_x:
            raise ValueError('jij_calc_x is not a DMI calculation (i.e. used old solver)')
    if dmimode and jij_calc_y_retrieved is not None:
        jij_calc_y = jij_calc_y_retrieved.get_incoming(node_class=KkrCalculation).first().node
        if verbose:
            print('jij(y) calculation:', jij_calc_y.uuid)
        jijs_shells_y, dmimode_y = get_jijs_shells(jij_calc_y)
        if verbose:
            print('shape jij_shells_y:', len(jijs_shells_y), dmimode_y)
        if not dmimode_y:
            raise ValueError('jij_calc_y is not a DMI calculation (i.e. used old solver)')
    # consistency check
    if dmimode and not (dmimode_x and dmimode_y):
        raise ValueError('Found dmimode but not all calculations (x,y and z) were given correctly')

    # read the shells information
    shells = get_shells(jij_calc, verbose)
    # print('shells', shells)

    # expand shells data to a structure that spirit can understand (i.e. take all pairs)
    if not dmimode:
        if verbose:
            print('expand isotropic Jijs')
        jijs_expanded, positions_expanded = expand_jijs_iso(jij_calc, jijs_shells_z, shells, alat)
    else:
        if verbose:
            print('expand anisotropic Jijs, Dij etc')
        jijs_expanded, positions_expanded = expand_jijs_dmi(
            jijs_shells_x, jijs_shells_y, jijs_shells_z, shells, cell, get_sites(structure), alat, verbose=verbose
        )

    # sort arrays
    isort = np.lexsort(jijs_expanded[:, :5][:, ::-1].transpose())
    jijs_expanded = jijs_expanded[isort]
    positions_expanded = positions_expanded[isort]

    # create an auxiliary structure that contains only the sites which are used in the Jij step
    # (i.e. we drop all sites where we don't have couplings)
    struc_jij_sites, mappings_back, mu_s = get_jij_structure(structure, jijs_expanded, jij_calc)

    if verbose:
        print(f'reduced structure has {len(struc_jij_sites.sites)} sites')

    # change i,j indices to match smaller structure
    # and correct sign if mu_i and mu_j are oriented antiparallel
    for ii, iatom in enumerate(jijs_expanded[:, 0]):
        jatom = jijs_expanded[ii, 1]
        jijs_expanded[ii, 0] = mappings_back[int(iatom)]
        jijs_expanded[ii, 1] = mappings_back[int(jatom)]
        # correct sign
        smu_i = np.sign(mu_s[int(jijs_expanded[ii, 0])])
        smu_j = np.sign(mu_s[int(jijs_expanded[ii, 1])])
        sign_factor = smu_i * smu_j
        jijs_expanded[ii, 5] *= sign_factor  # change sign of Jij
        jijs_expanded[ii, 6:9] *= sign_factor  # change sign of Dij

    # now collect the outputs in AiiDA Array objects
    jij_data = ArrayData()
    # jij_data.set_array('Jij_shells', jijs_shells_z)
    # if dmimode:
    #     # also add x and y shells output
    #     jij_data.set_array('Jij_shells_x', jijs_shells_x)
    #     jij_data.set_array('Jij_shells_y', jijs_shells_y)
    jij_data.set_array('Jij_expanded', jijs_expanded)
    jij_data.set_array('positions_expanded', positions_expanded)
    # add description to extras
    jij_data.extras['array_descriptions'] = {
        #         'Jij_shells': """Jij output in the shells that KKR found.
        #     The format differs for the type of calculation:
        #     - isotropic exchange constants only (old solver). The columns refer to:
        #       * [0] |Rij| in units of the lattice constant
        #       * [1] Jij in Ryd
        #       * [2] shell-number
        #       * [3] atom type of atom j
        #     - full exchange tensor (new solver). The columns refer to:
        #       * [0] ∣Rij∣ in units of the lattice constant
        #       * [1] Jij (isotropic part) in Ryd
        #       * [2] Dij (anti-symmetric DMI part) in Ryd
        #       * [3] Sij (diagonal traceless part) in Ryd
        #       * [4] Aij (off-diagonal symmetric part) in Ryd
        #       * [5-7] Rj - Ri (3 component vector)
        #       * [8] atom type of atom j

        #     if the full exchange tensor is calculated the Jij_shells_x and Jij_shells_y also exist""",
        'Jij_expanded':
        'i, j, da, db, dc, Jij (meV) [, Dij vector (x, y, z in meV), full Jij tensor (xx, xy, xz, yx, yy, yz, zx, zy, zz in meV)]',
        'positions_expanded': 'x, y, z (Ang.) positions of all pairs in Jij_expanded',
    }

    # add extras to generated structure for quick access
    struc_jij_sites.extras['mappings_ij'] = mappings_back
    struc_jij_sites.extras['uuid_struc_orig'] = structure.uuid
    struc_jij_sites.extras['uuid_jij_data'] = jij_data.uuid

    # return dict (link_label: node)
    return {'jij_data': jij_data, 'structure_jij_sites': struc_jij_sites}


def parse_jij_calc(jij_calc, jij_calc_x=None, jij_calc_y=None, verbose=False):
    """
    Parse outcome of Jij calculation and return jij_data and structure_jij_sites.
    Calculate not only Jij but also Dij vector if 3 calculations for m ||z, m||x and m||y are given.

    :param jij_calc: KkrCalculation with Jij inputs for m || z
    :param jij_calc_x: optional, KkrCalculation with Jij inputs for m || x
    :param jij_calc_y: optional, KkrCalculation with Jij inputs for m || y
    :param verbose: boolean that activate verbose writeout during parsing

    :return jij_data: ArrayData that contains the expanded Jij's (see 'array_descriptions' extra of the node for more details)
    :return struc_jij_sites: the reduced structure that contains onlt the atoms which have Jij couplings
        (comes from the input of the KkrCalculation). The mappings to the original structure (and their i,j indices) is given as an extra.
    """

    if not jij_calc.is_finished_ok:
        raise ValueError('Jij (z) calculation not finished ok (yet)!')
    if jij_calc_x is not None and not jij_calc_x.is_finished_ok:
        raise ValueError('Jij (x) calculation not finished ok (yet)!')
    if jij_calc_y is not None and not jij_calc_y.is_finished_ok:
        raise ValueError('Jij (y) calculation not finished ok (yet)!')

    if 'parsed_jij_data' in jij_calc.extras:
        # load from extras
        if verbose:
            print('found parsed jij data, load it now')
        jij_data_uuid = jij_calc.extras['parsed_jij_data']['jij_data_uuid']
        struc_jij_sites_uuid = jij_calc.extras['parsed_jij_data']['struc_jij_sites_uuid']
        jij_data = load_node(jij_data_uuid)
        structure_jij_sites = load_node(struc_jij_sites_uuid)
    else:
        # no old data found, do parsing here
        # the following calcfunction keeps the data provenance by taking the retrieved folders as inputs
        if jij_calc_x is None:
            jij_calc_x_retrieved = None
        else:
            jij_calc_x_retrieved = jij_calc_x.outputs.retrieved
        if jij_calc_y is None:
            jij_calc_y_retrieved = None
        else:
            jij_calc_y_retrieved = jij_calc_y.outputs.retrieved
        jij_parsed = parse_jij_calc_data(
            jij_calc.outputs.retrieved, jij_calc_x_retrieved, jij_calc_y_retrieved, verbose=Bool(verbose)
        )
        jij_data, structure_jij_sites = jij_parsed['jij_data'], jij_parsed['structure_jij_sites']

        # then save as extra to be able to load it in the next run
        if verbose:
            print('finished parsing jij data, save it as extra now')
        jij_calc.set_extra(
            'parsed_jij_data', {
                'jij_data_uuid': jij_data.uuid,
                'struc_jij_sites_uuid': structure_jij_sites.uuid
            }
        )

    # finally return result
    return jij_data, structure_jij_sites


def load_or_submit_cpa_jijcalcs(
    scf_cpa_wf,
    JIJSITEI=None,
    JIJSITEJ=None,
    tempr=400.,
    rclustz=0.9,
    kmesh=[100, 100, 100],
    NATOMIMPD=500,
    NSHELD=2000,
    JIJRAD=2.0,
    options=None,
    uuid_x=None,
    uuid_y=None,
    uuid_z=None
):
    """Load or submit Jij calculation for CPA"""

    scf_remote = scf_cpa_wf.outputs.last_RemoteData
    last_calc = scf_remote.get_incoming(node_class=KkrCalculation).first().node

    builder = last_calc.get_builder_restart()
    builder.parent_folder = scf_remote

    # set Jij parameters
    para_Jij = {k: v for k, v in scf_cpa_wf.outputs.last_InputParameters.get_dict().items() if v}
    para_Jij['TEMPR'] = tempr  # slightly reduce temperature
    para_Jij['RCLUSTZ'] = rclustz  # increase cluster radius
    para_Jij['BZDIVIDE'] = kmesh  # increase k-points
    para_Jij['NSTEPS'] = 1  # one-shot
    para_Jij['NATOMIMPD'] = NATOMIMPD  # array dimension
    para_Jij['NSHELD'] = NSHELD  # array dimension
    para_Jij['KPOIBZ'] = np.product(kmesh)  # array dimension
    # add 'XCPL' runopt to list of runopts (activates Jij calculation)
    runopts = para_Jij.get('RUNOPT', [])
    runopts.append('XCPL    ')
    para_Jij['RUNOPT'] = runopts
    # set Jij parameters
    # i and j index for Jij calculation in internal units
    # uses site index (i.e. needs to be <=10)
    if JIJSITEI is not None:
        para_Jij['JIJSITEI'] = JIJSITEI
        if JIJSITEJ is None:
            JIJSITEJ = JIJSITEI
    if JIJSITEJ is not None:
        para_Jij['JIJSITEJ'] = JIJSITEJ
    para_Jij['JIJRAD'] = JIJRAD  # radius in lattice constants up to which the Jijs are calculated

    builder.parameters = Dict(para_Jij)

    # starting angles for 3 directions, needed to extract full Jij tensor

    Nsites = len(get_sites(find_parent_structure(last_calc)))

    init_angles_x = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [90.0 for i in range(Nsites)],
        'phi': [0.0 for i in range(Nsites)],
    })

    init_angles_y = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [90.0 for i in range(Nsites)],
        'phi': [90.0 for i in range(Nsites)],
    })

    init_angles_z = Dict({
        'fix_dir': [True for i in range(Nsites)],
        'theta': [0.0 for i in range(Nsites)],
        'phi': [0.0 for i in range(Nsites)],
    })

    # submit m||z calculation
    builder.initial_noco_angles = init_angles_z
    builder.metadata.label = 'Jij_z'
    if options is not None:
        builder.metadata.options = options
    if uuid_z is not None:
        calc_Jij_z = load_node(uuid_z)
    else:
        calc_Jij_z = submit(builder)
        print('submitted z', calc_Jij_z)

    builder.initial_noco_angles = init_angles_y
    builder.metadata.label = 'Jij_y'
    if options is not None:
        builder.metadata.options = options
    if uuid_y is not None:
        calc_Jij_y = load_node(uuid_y)
    else:
        calc_Jij_y = submit(builder)
        print('submitted y', calc_Jij_y)

    builder.initial_noco_angles = init_angles_x
    builder.metadata.label = 'Jij_x'
    if options is not None:
        builder.metadata.options = options
    if uuid_x is not None:
        calc_Jij_x = load_node(uuid_x)
    else:
        calc_Jij_x = submit(builder)
        print('submitted x', calc_Jij_x)

    return calc_Jij_x, calc_Jij_y, calc_Jij_z



# -*- coding: utf-8 -*-

from aiida.parsers.parser import Parser
from aiida.orm import Dict
from aiida_kkr.calculations.voro import VoronoiCalculation
from aiida.common.exceptions import InputValidationError, NotExistent
from masci_tools.io.parsers.voroparser_functions import parse_voronoi_output
from aiida_kkr.tools.context import open_context_to_stack, open_files_in_context
import os
from contextlib import ExitStack

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.4.0'
__contributors__ = ('Jens Broeder', 'Philipp Rüßmann')


class VoronoiParser(Parser):
    """
    Parser class for parsing output of voronoi code..
    """

    def __init__(self, calc):
        """
        Initialize the instance of Voronoi_Parser
        """
        # these files should be present after success of voronoi
        self._default_files = {
            'outfile': VoronoiCalculation._OUTPUT_FILE_NAME,
            'atominfo': VoronoiCalculation._ATOMINFO,
            'radii': VoronoiCalculation._RADII
        }

        self._ParserVersion = __version__

        # reuse init of base class
        super(VoronoiParser, self).__init__(calc)

    # pylint: disable=protected-access
    def parse(self, debug=False, **kwargs):
        """
        Parse output data folder, store results in database.

        :param retrieved: a dictionary of retrieved nodes, where
          the key is the link name
        :returns: nothing if everything is fine or an exit code defined in the voronoi calculation class
        """

        success = False
        node_list = ()

        # Get retrieved folders
        try:
            out_folder = self.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = out_folder.list_object_names()

        # we need at least the output file name as defined in calcs.py
        if VoronoiCalculation._OUTPUT_FILE_NAME not in list_of_files:
            self.logger.error(f"Output file '{VoronoiCalculation._OUTPUT_FILE_NAME}' not found")
            return self.exit_codes.ERROR_NO_OUTPUT_FILE

        # Parse voronoi output, results that are stored in database are in out_dict

        # get path to files and catch errors if files are not present
        file_errors = []
        potfile_name = VoronoiCalculation._OUT_POTENTIAL_voronoi
        if potfile_name not in out_folder.list_object_names():
            # cover case where potfile is overwritten from input to voronoi calculation
            if VoronoiCalculation._POTENTIAL_IN_OVERWRITE in out_folder.list_object_names():
                potfile_name = VoronoiCalculation._POTENTIAL_IN_OVERWRITE
            else:
                file_errors.append(
                    'Critical error! Neither potfile {}  not {} '
                    'was found'.format(potfile_name, VoronoiCalculation._POTENTIAL_IN_OVERWRITE)
                )
                potfile_name = None

        outfile_name = VoronoiCalculation._OUTPUT_FILE_NAME
        if outfile_name not in out_folder.list_object_names():
            file_errors.append(f'Critical error! outfile not found {outfile_name}')
            outfile_name = None

        atominfo_name = VoronoiCalculation._ATOMINFO
        if atominfo_name not in out_folder.list_object_names():
            file_errors.append(f'Critical error! atominfo not found {atominfo_name}')
            atominfo_name = None

        radii_name = VoronoiCalculation._RADII
        if radii_name not in out_folder.list_object_names():
            file_errors.append(f'Critical error! radii not found {radii_name}')
            radii_name = None

        inputfile_name = VoronoiCalculation._INPUT_FILE_NAME
        if inputfile_name not in out_folder.list_object_names():
            file_errors.append(f'Critical error! inputfile not found {inputfile_name}')
            inputfile_name = None

        # initialize out_dict and parse output files
        out_dict = {'parser_version': self._ParserVersion}
        out_dict['calculation_plugin_version'] = VoronoiCalculation._CALCULATION_PLUGIN_VERSION
        # TODO add job description, compound name, calculation title

        with ExitStack() as stack:
            # open files only if they exist
            # openend files are added to the context manager stack
            # so that at the end of the parsing all files are closed
            # when the context manager is exited
            (potfile, outfile, atominfo, radii, inputfile) = open_files_in_context(
                stack, out_folder, potfile_name, outfile_name, atominfo_name, radii_name, inputfile_name
            )
            success, msg_list, out_dict = parse_voronoi_output(
                out_dict, outfile, potfile, atominfo, radii, inputfile, debug=debug
            )

        # add file open errors to parser output of error messages
        for f_err in file_errors:
            msg_list.append(f_err)
        out_dict['parser_errors'] = msg_list

        # create output node and link
        self.out('output_parameters', Dict(dict=out_dict))

        if not success:
            return self.exit_codes.ERROR_VORONOI_PARSING_FAILED



# -*- coding: utf-8 -*-
"""
This contains code snippets for making a supercell in one direction. It can be run repeatedly in various directions, in order to generate a larger supercell.
"""

#Check if accuracy is high enough!
from pymatgen.core import Lattice, Structure
import numpy as np


def make_super_cell(structure, scale_bravais_c, bravais_axis=2):
    matrix0 = np.array(np.asarray(structure.lattice.matrix), dtype=np.float128)
    coords = np.array(np.asarray(structure.cart_coords), dtype=np.float128)
    atomic_numbers = list(structure.atomic_numbers)
    new_atomic_numbers = list(structure.atomic_numbers)
    new_coords = np.copy(coords)
    for k in np.arange(1, scale_bravais_c):
        # print(np.shape(coords))
        new_coords_block = np.zeros(np.shape(coords))
        for j in np.arange(np.shape(coords)[0]):
            new_coords_block[j, :] = coords[j, :] + k * matrix0[bravais_axis, :]

        new_coords = np.concatenate((new_coords, new_coords_block), axis=0)
        new_atomic_numbers = new_atomic_numbers + atomic_numbers

    # print(new_coords)
    # print(new_atomic_numbers)
    matrix = np.copy(matrix0)
    matrix[2, :] = scale_bravais_c * matrix0[2, :]
    latt = Lattice(matrix)
    struc = Structure(latt, new_atomic_numbers, new_coords, coords_are_cartesian=True)
    return struc



"""
Helper tools to find parent calculations or structures.
"""

from aiida.common.exceptions import NotExistent


def get_struc(parent_calc):
    """
    Get structure from a parent_folder (result of a calculation, typically a remote folder)
    """
    return parent_calc.inputs.structure


def has_struc(parent_folder):
    """
    Check if parent_folder has structure information in its input
    """
    return 'structure' in parent_folder.base.links.get_incoming().all_link_labels()


def get_remote(parent_folder):
    """
    get remote_folder from input if parent_folder is not already a remote folder
    """
    parent_folder_tmp0 = parent_folder
    try:
        parent_folder_tmp = parent_folder_tmp0.base.links.get_incoming().get_node_by_label('remote_folder')
    except NotExistent:
        try:
            # check if GFhost_folder is there, this is the case for a KkrimpCalculation
            parent_folder_tmp = parent_folder_tmp0.base.links.get_incoming().get_node_by_label('GFhost_folder')
        except NotExistent:
            parent_folder_tmp = parent_folder_tmp0
    return parent_folder_tmp


def get_parent(input_folder):
    """
    Get the parent folder of the calculation. If no parent was found return input folder
    """
    input_folder_tmp0 = input_folder

    # first option: parent_calc_folder (KkrimpCalculation)
    try:
        parent_folder_tmp = input_folder_tmp0.base.links.get_incoming().get_node_by_label('parent_calc_folder')
        return_input = False
    except NotExistent:
        return_input = True

    # second option: parent_folder (KkrCalculation)
    try:
        parent_folder_tmp = input_folder_tmp0.base.links.get_incoming().get_node_by_label('parent_folder')
        return_input = False
    except NotExistent:
        return_input = return_input & True

    # third option: parent_KKR option (special mode of VoronoiCalculation)
    try:
        parent_folder_tmp = input_folder_tmp0.base.links.get_incoming().get_node_by_label('parent_KKR')
        return_input = False
    except NotExistent:
        return_input = return_input & True

    if return_input:
        parent_folder_tmp = input_folder_tmp0

    return parent_folder_tmp


def find_parent_structure(parent_folder):
    """
    Find the Structure node recuresively in chain of parent calculations (structure node is input to voronoi calculation)
    """
    iiter = 0
    Nmaxiter = 1000
    parent_folder_tmp = get_remote(parent_folder)
    while not has_struc(parent_folder_tmp) and iiter < Nmaxiter:
        parent_folder_tmp = get_remote(get_parent(parent_folder_tmp))
        iiter += 1
        if iiter % 200 == 0:
            print(
                'Warning: find_parent_structure takes quite long (already searched {} ancestors). Stop after {}'.format(
                    iiter, Nmaxiter
                )
            )
    if has_struc(parent_folder_tmp):
        struc = get_struc(parent_folder_tmp)
        return struc, parent_folder_tmp
    else:
        raise ValueError('structure not found')


def get_calc_from_remote(calc_remote):
    """
    Get the parent calculation from a RemoteData
    """
    from aiida.orm import CalcJobNode
    from aiida.orm import RemoteData

    if not isinstance(calc_remote, RemoteData):
        raise ValueError('input node is not a RemoteData folder')

    parents = calc_remote.base.links.get_incoming(node_class=CalcJobNode).all()
    if len(parents) != 1:
        raise ValueError('Parent is not unique!')

    return parents[0].node



# -*- coding: utf-8 -*-
"""
Calcfucntion that extracts the nonco angles from the output of a KkrCalculation
"""

from aiida.orm import Dict
from aiida.engine import calcfunction
from aiida_kkr.calculations import KkrCalculation
from numpy import sqrt, loadtxt


@calcfunction
def extract_noco_angles(**kwargs):
    """
    Extract noco angles from retrieved nonco_angles_out.dat files and save as Dict node which can be used as initial values for the next KkrCalculation.
    New angles are compared to old angles and if they are closer thanfix_dir_threshold they are not allowed to change anymore
    """
    # for comparison read previous theta and phi values and the threshold after which the moments are kept fixed
    noco_angles_old = kwargs['old_noco_angles'].get_dict()
    natom = len(noco_angles_old['phi'])
    noco_angles_old = [
        [noco_angles_old['theta'][i], noco_angles_old['phi'][i], noco_angles_old['fix_dir'][i]] for i in range(natom)
    ]
    fix_dir_threshold = kwargs['fix_dir_threshold'].value

    last_retrieved = kwargs['last_retrieved']
    noco_out_name = KkrCalculation._NONCO_ANGLES_OUT
    if noco_out_name in last_retrieved.list_object_names():
        with last_retrieved.open(noco_out_name) as noco_file:
            noco_angles_new = loadtxt(noco_file, usecols=[0, 1])
            # check if theta and phi change less than fix_dir_threshold
            fix_dir = [
                sqrt((noco_angles_new[i][0] - noco_angles_old[i][0])**2 +
                     (noco_angles_new[i][1] - noco_angles_old[i][1])**2) < fix_dir_threshold for i in range(natom)
            ]
            new_initial_noco_angles = Dict({
                'theta': list(noco_angles_new[:, 0]),
                'phi': list(noco_angles_new[:, 1]),
                # convert from numpy.bool_ to standard python bool, otherwise this is not json serializable and cannot be stored
                'fix_dir': [bool(i) for i in fix_dir]
            })
        return new_initial_noco_angles
